/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import teajay.bytecode.*;
import teajay.bytecode.attributes.code.*;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;

/**
 * Example of the high level API in use.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class NinetyNineBottlesOfBeerGenerator {
   
    public static void main(String... args) throws ByteCodeException, IOException, ClassNotFoundException {
        ClassFile bottles = new ClassFile("NinetyNineBottlesOfBeer", ClassAccessFlag.SUPER, ClassAccessFlag.PUBLIC, ClassAccessFlag.FINAL);
        
        MethodInfo main = new MethodInfo("main",
                new MethodDescriptor(FieldDescriptor.getVoid(), FieldDescriptor.fromClass(String[].class)), bottles.getThis(),
                MethodAccessFlag.PUBLIC, MethodAccessFlag.STATIC);
        bottles.addMethod(main);
        
        CodeAttribute code = main.getCode();

        generateMain(code);
        
        //Add private Constructor
        MethodInfo init = new MethodInfo("<init>", new MethodDescriptor(FieldDescriptor.getVoid()), bottles.getThis(), MethodAccessFlag.PRIVATE);
        bottles.addMethod(init);
        //Mandatory super constructor call
        JVMInstructionList lst = init.getCode().getInstructionList();
        lst.add(new ALoad(0));
        lst.add(new InvokeSpecial(MethodRefInfo.make("java/lang/Object", "<init>", "void")));
        
        lst.add(new Return()); //void return
        
        
        File dest = new File("/home/michi/NinetyNineBottlesOfBeer.class");
        try (DataOutputStream out = new DataOutputStream(new FileOutputStream(dest))) {
            bottles.write(out);
        } catch (ByteCodeException e) {
            e.printStackTrace(System.err);
            dest.delete();
        }
    }

    private static void generateMain(CodeAttribute code) throws ByteCodeException {
        StackMapTable table = code.getStackMapTable();
        LocalVariableTable lt = new LocalVariableTable();
        code.addAttribute(lt);

        MethodDescriptor parseIntDescr = new MethodDescriptor(FieldDescriptor.fromBaseType(BaseType.Int), FieldDescriptor.fromClass(String.class));
        MethodRefInfo parseInt = new MethodRefInfo(new ClassInfo("java/lang/Integer"), new NameAndTypeInfo<>("parseInt", parseIntDescr));

        FieldDescriptor printStream = FieldDescriptor.fromString("java/io/PrintStream");

        FieldRefInfo systemOut = new FieldRefInfo(new ClassInfo("java/lang/System"), new NameAndTypeInfo<>("out", printStream));

        FieldDescriptor voidDescr = FieldDescriptor.fromClass(void.class);
        
        MethodRefInfo printString = new MethodRefInfo(new ClassInfo(printStream), new NameAndTypeInfo<>("print", 
                new MethodDescriptor(voidDescr, FieldDescriptor.fromClass(String.class))));

        MethodRefInfo printInt = new MethodRefInfo(new ClassInfo(printStream), new NameAndTypeInfo<>("print", 
                new MethodDescriptor(voidDescr, FieldDescriptor.fromClass(int.class))));

        JVMInstructionList lst = code.getInstructionList();

        Label defaultLength = new Label();
        JVMInstruction jumpback = new GetStatic(systemOut);
        JVMInstruction voidreturn = new Return();

        JVMInstruction startPC = lst.add(new ALoad(0)); //Load String[] args
        lst.add(new ArrayLength());
        lst.add(new IConst(1));
        lst.add(new IfICmp(IMode.lt, defaultLength)); //If no argument exists use default 99

        //load index 0 of String[] args
        lst.add(new ALoad(0));
        lst.add(new IConst(0));
        lst.add(new AALoad());
        //Parse the first command line parameter to an integer
        lst.add(new InvokeStatic(parseInt));
        //Store the parsed argument in local variable 1
        lt.addMapping(lst.add(new IStore(1)), voidreturn, "bottlesOfBeer", FieldDescriptor.fromBaseType(BaseType.Int), 1);
        lst.add(new Goto(jumpback));

        table.addSameFrame(lst.add(defaultLength, new IConst(99))); //bipush 99
        //table.addSameFrame(defaultLength);
        //Store the integer 99 in local variable 1
        lst.add(new IStore(1));
        
        //Define jump marks
        PlaceholderInstruction iinc = new PlaceholderInstruction();
        JVMInstruction getstatic1 = new GetStatic(systemOut);
        JVMInstruction getstatic2 = new GetStatic(systemOut);
        JVMInstruction iload1 = new ILoad(1);
        JVMInstruction getstatic4 = new GetStatic(systemOut);
        //Loop start
        lst.add(jumpback);
        //We need a new stack frame
        AppendFrame frame = table.addAppendFrame(jumpback);
        frame.addLocal(VerificationType.makeInteger()); //only one integer is in use here

        //Print bottle count
        lst.add(new ILoad(1));
        lst.add(new InvokeVirtual(printInt));

        //Check for bottle count beeing 1
        lst.add(new ILoad(1));
        //lst.add(wide1, iload, new U2(1));
        lst.add(new IConst(1));
        lst.add(new IfICmp(IMode.eq, getstatic2));

        //Print default text
        lst.add(new GetStatic(systemOut));
        lst.add(new Ldc(" bottles of beer on the wall,\n"));
        lst.add(new InvokeVirtual(printString));
        //and jump over the special case
        lst.add(new Goto(getstatic1));

        //Special case: number of bottles is 1
        lst.add(getstatic2);
        table.addSameFrame(getstatic2);
        lst.add(new Ldc(" bottle of beer on the wall,\n"));
        lst.add(new InvokeVirtual(printString));

        //Print bottle count
        lst.add(getstatic1);
        table.addSameFrame(getstatic1);
        lst.add(new ILoad(1));
        lst.add(new InvokeVirtual(printInt));

        //Check for bottle count beeing 1
        lst.add(new ILoad(1));
        lst.add(new IConst(1));
        lst.add(new IfICmp(IMode.eq, getstatic4));

        //Print the second default text
        lst.add(new GetStatic(systemOut));
        lst.add(new Ldc(" bottles of beer.\n"));
        lst.add(new InvokeVirtual(printString));
        //and jump over the special case
        lst.add(new Goto(iload1));

        //Special case: number of bottles is 1
        lst.add(getstatic4);
        table.addSameFrame(getstatic4);
        lst.add(new Ldc(" bottle of beer.\n"));
        lst.add(new InvokeVirtual(printString));

        //Jump over the last sentence if all bottles are gone
        lst.add(iload1); //iload_1
        table.addSameFrame(iload1);

        lst.add(new If0(IMode.eq, iinc)); // == 0

        //Print the third default text
        lst.add(new GetStatic(systemOut));
        lst.add(new Ldc("Take one down, pass it around,\n"));
        lst.add(new InvokeVirtual(printString));

        //Decrement bottle count
        lst.add(iinc); //iinc, 1, -1
        table.addSameFrame(iinc);

        lst.add(new ILoad(1));
        //Check bottles for beeing greater equal 0
        lst.add(new If0(IMode.ge, jumpback)); // >= 0

        lst.add(voidreturn);
        
        JVMInstruction exceptionHandler = lst.add(new Pop());
        lst.add(new GetStatic(systemOut));
        
        FullFrame full = table.addFullFrame(exceptionHandler);
        ClassInfo nfe = ClassInfo.fromClass(NumberFormatException.class);
        full.addStack(VerificationType.valueOf(nfe));
        full.addLocal(VerificationType.valueOf(String[].class));
        
        lst.add(new Ldc("Expecting integer argument.\n"));
        lst.add(new InvokeVirtual(printString));
        
        JVMInstruction lastInstr = lst.add(new Return());
        lt.addMapping(startPC, lastInstr, "args", FieldDescriptor.fromClass(String[].class), 0);
        //Cover the hole code in an exception handler
        code.addExceptionHandler(startPC, voidreturn, exceptionHandler, nfe);
        
        iinc.setInstruction(new IInc(1, -1));
    }
}
