/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode;

import java.io.DataOutputStream;
import java.io.IOException;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * signed int
 * @author Michael Kußmann <michi@outx.de>
 */
public final class S4 implements Operand {

    public static final S4 ZERO = new S4();
    private final long value;

    private S4() {
        this.value = 0;
    }

    public S4(int value) {
        this.value = value;
    }

    public S4(long value) throws ByteCodeException {
        if (value < Integer.MIN_VALUE || value > Integer.MAX_VALUE) {
            throw new ByteCodeException("signed int (S4) out of bounds: " + value);
        }
        this.value = (int) value;
    }

    public final long getValue() {
        return value;
    }

    @Override
    public long getAddress(ConstantPool cpool) {
        return value;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException {
        out.writeInt((int) value);
    }

    @Override
    public void writeIndex(DataOutputStream out, ConstantPool cpool) throws IOException {
        this.write(out, cpool);
    }

    @Override
    public OpCode.OpSize getDefaultLengthAsOperand() {
        return OpCode.OpSize.U4;
    }

    @Override
    public String toString() {
        return "S4: " + value;
    }

    @Override
    public boolean equals(Object obj) {
        if (obj instanceof S4) {
            S4 other = (S4) obj;
            return other.value == value;
        }
        return false;
    }

    @Override
    public int hashCode() {
        int hash = 3;
        hash = 11 * hash + (int) (this.value ^ (this.value >>> 32));
        return hash;
    }
}
