/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;

/**
 * These flags have the same value in all contexts.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public enum VisibilityAccessFlags {

    PACKAGE_PRIVATE(0x0000),
    PUBLIC(0x0001),
    PRIVATE(0x0002),
    PROTECTED(0x0004);
    public final int intValue;

    private VisibilityAccessFlags(int value) {
        this.intValue = value;
    }

    public static boolean checkPackagePrivate(int bits) {
        return ((PUBLIC.intValue | PRIVATE.intValue | PROTECTED.intValue) & bits) == 0;
    }

    @SafeVarargs
    @SuppressWarnings({"unchecked", "varargs"})
    private static <T extends AccessFlag> boolean checkPackagePrivate(T... flags) {
        return checkPackagePrivate(Arrays.asList(flags));
    }

    private static <T extends AccessFlag> boolean checkPackagePrivate(Collection<T> flags) {
        for (T flag : flags) {
            switch (flag.getIntValue()) {
                case 0x0000:
                case 0x0001:
                case 0x0002:
                case 0x0004:
                    return false;
            }
        }
        return true;
    }

    private static <T extends AccessFlag> T getPackagePrivate(Class<T> type) {
        T[] c = type.getEnumConstants();
        for (T flag : c) {
            if (flag.getIntValue() == PACKAGE_PRIVATE.intValue) {
                return flag;
            }
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    public static <T extends AccessFlag> T[] handlePackagePrivate(T[] accessFlags, Class<T> type) {
        if (VisibilityAccessFlags.checkPackagePrivate(accessFlags)) {
            T[] temp = (T[]) Array.newInstance(type, accessFlags.length + 1);
            System.arraycopy(accessFlags, 0, temp, 0, accessFlags.length);
            T pp = getPackagePrivate(type);
            if (pp != null) {
                temp[accessFlags.length] = pp;
            } else {
                throw new RuntimeException("Invalid invocation of handlePackagePrivate.");
            }
            return temp;
        } else {
            return accessFlags;
        }
    }

    public static <T extends AccessFlag> Collection<T> handlePackagePrivate(Collection<T> accessFlags, Class<T> type) {
        Collection<T> temp = new HashSet<>();
        temp.addAll(accessFlags);
        if (VisibilityAccessFlags.checkPackagePrivate(accessFlags)) {
            T pp = getPackagePrivate(type);
            if (pp != null) {
                temp.add(pp);
            } else {
                throw new RuntimeException("Invalid invocation of handlePackagePrivate.");
            }
        }
        return temp;
    }
}
