/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.annotations;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class Annotation implements ConstantPoolAddable, ByteCodeWritable, Iterable<Entry<Utf8Info, ElementValue>> {
     
    private final FieldDescriptor type;
    private final Map<Utf8Info, ElementValue> values;
    
    public Annotation(FieldDescriptor type) {
        this.type = type;
        this.values = new HashMap<>();
    }
    
    int getLength() {
        int len = 2 + 2;
        for (ElementValue entry : values.values()) {
            len += 2 + entry.getLength();
        }
        return len;
    }
    
    public FieldDescriptor getType() {
        return type;
    }
    
    public void addElementValue(Utf8Info name, ElementValue value) {
        values.put(name, value);
    }
    
    public void addElementValue(String name, ElementValue value) throws UTFOutOfBoundsException {
        addElementValue(new Utf8Info(name), value);
    }
    
    public ElementValue getElementValue(String name) {
        return getElementValue(new Utf8Info(name));
    }
    
    public ElementValue getElementValue(Utf8Info name) {
        return values.get(name);
    }
    
    public void removeElementValue(String name) {
        values.remove(new Utf8Info(name));
    }
    
    public void removeElementValue(Utf8Info name) {
        values.remove(name);
    }
    
    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        buf.append("Annotation type=");
        buf.append(type);
        buf.append(" values = [");
        for (Entry<Utf8Info, ElementValue> entry : values.entrySet()) {
            buf.append(entry.getKey().getValue());
            buf.append(" : " );
            buf.append(entry.getValue());
            buf.append(", ");
        }
        buf.deleteCharAt(buf.length() - 1);
        buf.deleteCharAt(buf.length() - 1);
        buf.append("]");
        return buf.toString();
    }
    
    @Override
    public Iterator<Entry<Utf8Info, ElementValue>> iterator() {
        return values.entrySet().iterator();
    }

    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        type.add(pool);
        for (Entry<Utf8Info, ElementValue> entry : values.entrySet()) {
            entry.getKey().add(pool);
            entry.getValue().add(pool);
        }
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        type.writeIndex(out, cpool);
        new U2(values.size()).write(out, cpool);
        for (Entry<Utf8Info, ElementValue> entry : values.entrySet()) {
            entry.getKey().writeIndex(out, cpool);
            entry.getValue().write(out, cpool);
        }
    }
    
    public static Annotation read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        FieldDescriptor type = ccpt(cpool.get(in.readUnsignedShort()), FieldDescriptor.class, cpool);
        int numPairs = in.readUnsignedShort();
        Annotation ann = new Annotation(type);
        for (int i = 0; i < numPairs; i++) {
            ann.addElementValue(ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool), ElementValue.read(in, cpool, major, minor));
        }
        return ann;
    }
}
