/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.Operand;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 * This is a method's attribute featuring the jvm byte code implementing the
 * method.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class CodeAttribute extends AttributeInfo {
    public static final String NAME = "Code";

    private final U2 maxLocals;
    private final U2 maxStack;
    private final JVMInstructionList instructionList;
    private final Collection<ExceptionTableEntry> exceptionTable;
    private final Map<String, AttributeInfo> attributes;
    private final StackMapTable stackMap;
   
    /**
     * Creates a new Code attribute.
     *
     * @param maxStack if != null overrides the max stack autodetection!
     * @param maxLocals if != null overrides the max locals autodetection!
     * @param firstFrame The implicit first frame generated by the methods argument list
     * @param instructionList the code implementing this method
     */
    private CodeAttribute(U2 maxStack, U2 maxLocals, FullFrame firstFrame) {
        super(new Utf8Info(NAME));
        this.maxLocals = maxLocals;
        this.exceptionTable = new ArrayList<>();
        this.attributes = new HashMap<>();
        this.maxStack = maxStack;
        this.stackMap = new StackMapTable(firstFrame);
        this.instructionList = new JVMInstructionList(stackMap);

        List<VerificationType> lst = firstFrame.getLocals();
        int current = lst.size();
        for (VerificationType type : lst) {
            //long and double take up two slots
            switch (type.getTag()) {
                case Long:
                case Double:
                    current++;
            }
        }
        instructionList.setInititalLocalsCount(current);
    }

    public CodeAttribute(FullFrame firstFrame) {
        this(null, null, firstFrame);
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        long attrLen = 0;
        for (AttributeInfo info : attributes.values()) {
            attrLen += info.getLength(cpool).getValue() + 6;
        }
        if (!stackMap.isEmpty()) {
            attrLen += stackMap.getLength(cpool).getValue() + 6;
        }
        return new U4(2 + 2 + 4 + instructionList.getLength(cpool).getValue() + 2 + 8 * exceptionTable.size() + 2 + attrLen);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (ExceptionTableEntry entry : exceptionTable) {
            entry.add(cpool);
        }
        for (AttributeInfo info : attributes.values()) {
            info.add(cpool);
        }
        if (!stackMap.isEmpty()) {
            stackMap.add(cpool);
        }
        
        instructionList.add(cpool);
    }

    public void addAttribute(AttributeInfo info) throws ByteCodeException {
        String key = info.getName().getValue();
        if (attributes.containsKey(key) || key.equals(StackMapTable.NAME)) {
            throw new ByteCodeException("Cannot have multiple attributes with the same name.");
        } else {
            attributes.put(key, info);
        }
    }

    public AttributeInfo getAttribute(String name) {
        return attributes.get(name);
    }

    public AttributeInfo getAttribute(Utf8Info name) {
        return attributes.get(name.getValue());
    }

    /**
     * Add an exception handler for the method's code. If <tt>catchType</tt> is
     * null the handler will catch everything. It is up to the user of the method to make sure
     * that all instruction are in correct order and added to this code attribute's instruction list by the time
     * of writing.
     *
     * @param startPC the start of the covered code block
     * @param endPC the end of the covered code block
     * @param handlerPC pointer to the start of the exception handling code
     * @param catchType the type of exception this handler will catch, use null for catch all
     */
    public void addExceptionHandler(JVMInstruction startPC, JVMInstruction endPC, JVMInstruction handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(Label startPC, Label endPC, Label handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(JVMInstruction startPC, Label endPC, Label handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(JVMInstruction startPC, JVMInstruction endPC, Label handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(JVMInstruction startPC, Label endPC, JVMInstruction handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(Label startPC, Label endPC, JVMInstruction handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }
    /**
     * @see #addExceptionHandler(teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.attributes.code.JVMInstruction, teajay.bytecode.cpool.ClassInfo) 
     * @param startPC
     * @param endPC
     * @param handlerPC
     * @param catchType 
     */
    public void addExceptionHandler(Label startPC, JVMInstruction endPC, Label handlerPC, ClassInfo catchType) {
        exceptionTable.add(new ExceptionTableEntry(startPC, endPC, handlerPC, catchType));
    }

    public Collection<ExceptionTableEntry> getExceptionTable() {
        return Collections.unmodifiableCollection(exceptionTable);
    }

    public StackMapTable getStackMapTable() {
        return stackMap;
    }

    public JVMInstructionList getInstructionList() {
        return instructionList;
    }

    public FullFrame getFirstFrame() {
        return stackMap.getFirstFrame();
    }
    
    public void removeExceptionHandlerByStartPC(JVMInstruction startPC) {
        removeExceptionHandlerByStartPC((Operand) startPC);
    }
    
    public void removeExceptionHandlerByStartPC(Label startPC) {
        removeExceptionHandlerByStartPC((Operand) startPC);
    }
    
    public void removeExceptionHandlerByEndPC(JVMInstruction endPC) {
        removeExceptionHandlerByEndPC((Operand) endPC);
    }
    
    public void removeExceptionHandlerByEndPC(Label endPC) {
        removeExceptionHandlerByEndPC((Operand) endPC);
    }
    
    public void removeExceptionHandlerByStartAndEndPC(JVMInstruction startPC, JVMInstruction endPC) {
        removeExceptionHandlerByStartAndEndPC((Operand) startPC, (Operand) endPC);
    }
    
    public void removeExceptionHandlerByStartAndEndPC(Label startPC, Label endPC) {
        removeExceptionHandlerByStartAndEndPC((Operand) startPC, (Operand) endPC);
    }
    
    public void removeExceptionHandlerByStartAndEndPC(Label startPC, JVMInstruction endPC) {
        removeExceptionHandlerByStartAndEndPC((Operand) startPC, (Operand) endPC);
    }
    
    public void removeExceptionHandlerByStartAndEndPC(JVMInstruction startPC, Label endPC) {
        removeExceptionHandlerByStartAndEndPC((Operand) startPC, (Operand) endPC);
    }
    
    private void removeExceptionHandlerByStartAndEndPC(Operand startPC, Operand endPC) {
        Iterator<ExceptionTableEntry> it = exceptionTable.iterator();
        boolean found = false;
        while (it.hasNext()) {
            if (it.next().getStartPC() == startPC && it.next().getEndPC() == endPC) {
                it.remove();
                found = true;
            }
        }
        if (!found) {
            throw new IllegalArgumentException("no such instruction: " + startPC);
        }
    }
    
    private void removeExceptionHandlerByStartPC(Operand startPC) {
        Iterator<ExceptionTableEntry> it = exceptionTable.iterator();
        boolean found = false;
        while (it.hasNext()) {
            if (it.next().getStartPC() == startPC) {
                it.remove();
                found = true;
            }
        }
        if (!found) {
            throw new IllegalArgumentException("no such instruction: " + startPC);
        }
    }
    
    private void removeExceptionHandlerByEndPC(Operand endPC) {
        Iterator<ExceptionTableEntry> it = exceptionTable.iterator();
        boolean found = false;
        while (it.hasNext()) {
            if (it.next().getEndPC() == endPC) {
                it.remove();
                found = true;
            }
        }
        if (!found) {
            throw new IllegalArgumentException("no such instruction: " + endPC);
        }
    }
    

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        instructionList.validate(cpool);
        stackMap.validate();

        super.write(out, cpool);

        if (maxStack != null) {
            maxStack.write(out, cpool);
        } else {
            instructionList.getMaxStackSize(stackMap, cpool).write(out, cpool);
        }
        if (maxLocals != null) {
            maxLocals.write(out, cpool);
        } else {
            instructionList.getMaxLocals(cpool).write(out, cpool);
        }
        instructionList.write(out, cpool);
        new U2(exceptionTable.size()).write(out, cpool);
        for (ExceptionTableEntry entry : exceptionTable) {
            entry.write(out, cpool);
        }
        int size = attributes.size();
        if (!stackMap.isEmpty()) {
            size++;
        }
        new U2(size).write(out, cpool);

        if (!stackMap.isEmpty()) {
            stackMap.write(out, cpool);
        }
        for (AttributeInfo info : attributes.values()) {
            info.write(out, cpool);
        }
    }
}
