/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import teajay.bytecode.U1;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class FullFrame extends StackVerifierStackMapFrame {
    public static final int FRAME_TYPE = 255;
    private final VerificationTypeList locals;
    
    FullFrame(JVMInstruction start) {
        super(start, new VerificationTypeList());
        this.locals = new VerificationTypeList();
    }
    /**
     * Creates a dummy frame.
     */
    public FullFrame() {
        super(null, new VerificationTypeList());
        this.locals = new VerificationTypeList();
    }
    
    FullFrame(List<VerificationType> locals, List<VerificationType> stack, JVMInstruction start) {
        super(start, new VerificationTypeList(stack));
        this.locals = new VerificationTypeList(locals);
    }
    
    @Override
    U1 getFrameType() throws ByteCodeException {
        return new U1(FRAME_TYPE);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        locals.add(cpool);
    }

    @Override
    U4 getLength() throws ByteCodeException {
        return new U4(1 + 2 + 2 + 2 + getStackTypeList().getLength().getValue() + locals.getLength().getValue());
    }
    
    public final VerificationTypeList getLocalTypeList() {
        return locals;
    }
    
    public void addLocal(VerificationType... type) throws ByteCodeException {
        addLocal(Arrays.asList(type));
    }
    
    public void addLocal(List<VerificationType> types) throws ByteCodeException {
        for (VerificationType type : types) {
            locals.add(type);
        }
    }
    
    @Deprecated
    public void addLocalTop() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Top));
    }
    @Deprecated
    public void addLocalDouble() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Double));
    }
    @Deprecated
    public void addLocalLong() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Long));
    }
    @Deprecated
    public void addLocalInteger() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Integer));
    }
    @Deprecated
    public void addLocalFloat() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Float));
    }
    @Deprecated
    public void addLocalNull() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.Null));
    }
    @Deprecated
    public void addLocalUninitializedThis() throws ByteCodeException {
        locals.add(new VerificationType(VariableInfo.UninitializedThis));
    }
    @Deprecated
    public void addLocalObject(ClassInfo className) throws ByteCodeException {
        locals.add(new VerificationType(className));
    }
    @Deprecated
    public void addLocalUninitialized(JVMInstruction newInstr) throws ByteCodeException {
        locals.add(new VerificationType(newInstr));
    }
    
    public void addLocalObjects(ClassInfo... classNames) throws ByteCodeException {
        for (ClassInfo className : classNames) {
            locals.add(VerificationType.valueOf(className));
        }
    }
    
    public void addLocalObjects(Collection<ClassInfo> classNames) throws ByteCodeException {
        for (ClassInfo className : classNames) {
            locals.add(VerificationType.valueOf(className));
        }
    }
    
    public void addLocalTypes(Collection<FieldDescriptor> types) throws ByteCodeException {
        for (FieldDescriptor type : types) {
            locals.add(VerificationType.valueOf(type));
        }
    }
      
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        offsetDelta().write(out, cpool);
        getLocalTypeList().getSize().write(out, cpool);
        getLocalTypeList().write(out, cpool);
        getStackTypeList().getSize().write(out, cpool);
        getStackTypeList().write(out, cpool);
    }

    @Override
    List<VerificationType> getStack() {
        return getStackTypeList().getList();
    }

    @Override
    List<VerificationType> getLocals() {
        return getLocalTypeList().getList();
    }
    
}
