/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import teajay.bytecode.OpCode;
import teajay.bytecode.Operand;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class Ldc extends JVMInstruction {
    private OpCode opcode;
    private final CPoolEntry entry;

    private Ldc(CPoolEntry entry) {
        this.entry = entry;
        switch (entry.getTag()) {
            case String:
            case Integer:
            case Float:
            case MethodType:
            case Class:
            case MethodHandle:
                this.opcode = OpCode.ldc;
                break;
            case Double:
            case Long:
                this.opcode = OpCode.ldc2_w;
                break;
            default:
                throw new RuntimeException("ldc: Invalid constant pool entry: " + entry);
        }
    }
    
    public Ldc(MethodHandleInfo<?> info) {
        this((CPoolEntry) info);
    }
    
    public Ldc(ClassInfo info) {
        this((CPoolEntry) info);
    }
    
    public Ldc(MethodTypeInfo info) {
        this((CPoolEntry) info);
    }
    
    public Ldc(StringInfo value) {
        this((CPoolEntry) value);
    }
    
    public Ldc(IntegerInfo value) {
        this((CPoolEntry) value);
    }
    
    public Ldc(FloatInfo value) {
        this((CPoolEntry) value);
    }
    
    public Ldc(DoubleInfo value) {
        this((CPoolEntry) value);
    }
    
    public Ldc(LongInfo value) {
        this((CPoolEntry) value);
    }
    
    public Ldc(String value) {
        this(new StringInfo(value));
    }
    
    public Ldc(int value) {
        this(new IntegerInfo(value));
    }
    
    public Ldc(float value) {
        this(new FloatInfo(value));
    }
    
    public Ldc(double value) {
        this(new DoubleInfo(value));
    }
    
    public Ldc(long value) {
        this(new LongInfo(value));
    }

    @Override
     int getLength() throws ByteCodeException {
        switch (opcode()) {
            case ldc:
                return 2;
            case ldc_w:
            case ldc2_w:
                return 3;
            default:
                throw new RuntimeException("Unreachable case reached.");
        }
    }

    @Override
    int getLocalModifier() {
        return 0;
    }

    @Override
    boolean validate(ConstantPool cpool) throws ByteCodeException {
        long address = entry.getAddress(cpool);
        if (opcode == OpCode.ldc && address > 0xFF) {
            this.opcode = OpCode.ldc_w;
            return false;
        } else {
            return true;
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        opcode.write(out, cpool);
        switch (opcode) {
            case ldc:
                entry.getCIndex(cpool).toU1Index().write(out, cpool);
                break;
            default:
                entry.getCIndex(cpool).write(out, cpool);
        }
    }

    @Override
    public OpCode opcode() throws ByteCodeException {
        return opcode;
    }

    @Override
    protected Collection<? extends Operand> operands() {
        return Arrays.asList(entry);
    }

    @Override
    int getPush() throws ByteCodeException {
        return opcode.num_push;
    }

    @Override
    int getPop() throws ByteCodeException {
        return opcode.num_pop;
    }
}
