/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.Operand;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */

public class LocalVariableTable extends AttributeInfo {
    public static final String NAME = "LocalVariableTable";
    
    private class LocalVariableEntry implements ByteCodeWritable {
        private final Operand startPC;
        private final Operand endPC;
        private final Utf8Info name;
        private final FieldDescriptor type;
        private final U2 localIndex;
        
        private LocalVariableEntry(Operand startPC, Operand endPC, Utf8Info name, FieldDescriptor type, U2 localIndex) {
            this.startPC = startPC;
            this.endPC = endPC;
            this.name = name;
            this.type = type;
            this.localIndex = localIndex;
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            long length = endPC.getAddress(cpool) - startPC.getAddress(cpool);
            if (length < 0 || length > 0xFFFF) {
                throw new ByteCodeException("LocalVariableTable: range out of bounds: " + length);
            }
            startPC.writeIndex(out, cpool);
            new U2(length).write(out, cpool);
            name.writeIndex(out, cpool);
            type.writeIndex(out, cpool);          
            localIndex.write(out, cpool);
        }
        
        public void add(ConstantPool cpool) throws ByteCodeException {
            type.add(cpool);
            name.add(cpool);
        }
    }
    
    List<LocalVariableEntry> locals = new ArrayList<>();

    public LocalVariableTable() {
        super(new Utf8Info(NAME));
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2 + locals.size() * 10);
    }
    
    public void addMapping(JVMInstruction startPC, JVMInstruction endPC, Utf8Info name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping((Operand) startPC, (Operand) endPC, name, type, localIndex);
    }
    
    private void addMapping(Operand startPC, Operand endPC, Utf8Info name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        if (localIndex < 0 || localIndex > 0xFFFF) {
            throw new ByteCodeException("LocalVariableTable: local variable index out of bounds (16bit unsigned short): " + localIndex);
        }
        locals.add(new LocalVariableEntry(startPC, endPC, name, type, new U2(localIndex)));
    }
    
    public void addMapping(JVMInstruction startPC, JVMInstruction endPC, String name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping(startPC, endPC, new Utf8Info(name), type, localIndex);
    }
    
    public void addMapping(JVMInstruction startPC, Label endPC, Utf8Info name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping((Operand) startPC, (Operand) endPC, name, type, localIndex);
    }
    
    public void addMapping(Label startPC, Label endPC, Utf8Info name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping((Operand) startPC, (Operand) endPC, name, type, localIndex);
    }
    
    public void addMapping(Label startPC, JVMInstruction endPC, Utf8Info name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping((Operand) startPC, (Operand) endPC, name, type, localIndex);
    }
    
    public void addMapping(Label startPC, JVMInstruction endPC, String name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping( (Operand) startPC, (Operand) endPC, new Utf8Info(name), type, localIndex);
    }

    public void addMapping(Label startPC, Label endPC, String name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping( (Operand) startPC, (Operand) endPC, new Utf8Info(name), type, localIndex);
    }
    
    public void addMapping(JVMInstruction startPC, Label endPC, String name, FieldDescriptor type, int localIndex) throws ByteCodeException {
        this.addMapping( (Operand) startPC, (Operand) endPC, new Utf8Info(name), type, localIndex);
    }
    
    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws ByteCodeException, IOException {
        super.write(out, cpool);
        new U2(locals.size()).write(out, cpool);
        for (LocalVariableEntry entry : locals) {
            entry.write(out, cpool);
        }
    }
    
    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (LocalVariableEntry entry : locals) {
            entry.add(cpool);
        }
    }
}