/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.code;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.util.UnmodifiableIterator;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class VerificationTypeList implements ByteCodeWritable, Iterable<VerificationType>, ConstantPoolAddable {

    private final int limit;
    private final List<VerificationType> types;

    VerificationTypeList(int limit) {
        this(new ArrayList<VerificationType>(), limit);
    }

    VerificationTypeList() {
        this(0xFFFF);
    }
    
    VerificationTypeList(List<VerificationType> types) {
        this(types, 0xFFFF);
    }
    
    VerificationTypeList(List<VerificationType> types, int limit) {
        this.types = types;
        this.limit = limit;
    }

    public U4 getLength() throws ByteCodeException {
        int len = 0;
        for (VerificationType type : types) {
            len += type.getLength().getValue();
        }
        return new U4(len);
    }

    public U2 getSize() throws ByteCodeException {
        return new U2(types.size());
    }

    public boolean isEmpty() {
        return types.isEmpty();
    }

    public void add(VerificationType type) throws ByteCodeException {
        if (types.size() < limit) {
            types.add(type);
        } else {
            throw new ByteCodeException("VerificationTypeList size exceeded: " + limit);
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        for (VerificationType type : types) {
            type.write(out, cpool);
        }
    }

    @Override
    public Iterator<VerificationType> iterator() {
        return new UnmodifiableIterator<>(types.iterator());
    }
    
    public List<VerificationType> getList() {
        return Collections.unmodifiableList(types);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        for (VerificationType type : this) {
            type.add(cpool);
        }
    }
}
