/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.innerclass;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import teajay.bytecode.U2;
import teajay.bytecode.U4;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 * Implementation of the EnclosingMethod attribute as stated in "The JavaTM
 * Virtual Machine Specification Version 7". Revision: 27.07.2012
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class EnclosingMethod extends AttributeInfo {

    public static final String NAME = "EnclosingMethod";
    private final ClassInfo innermostClass;
    private final NameAndTypeInfo<MethodDescriptor> enclosingMethod;

    private EnclosingMethod(ClassInfo innermostClass, NameAndTypeInfo<MethodDescriptor> enclosingMethod, Utf8Info name) {
        super(name);
        checkName(NAME, name);
        this.innermostClass = innermostClass;
        this.enclosingMethod = enclosingMethod;
    }

    /**
     * Creates a new enclosing method attribute. If there is no enclosing method
     * <tt>enclosingMethod</tt> may be null.
     *
     * @param innermostClass the innermost class that surrounds this class.
     * @param enclosingMethod the name and type of the enclosing method, may be
     * null
     */
    public EnclosingMethod(ClassInfo innermostClass, NameAndTypeInfo<MethodDescriptor> enclosingMethod) {
        this(innermostClass, enclosingMethod, new Utf8Info(NAME));
    }

    /**
     * Creates a new enclosing method attribute without a enclosing method.
     *
     * @see #EnclosingMethod(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.NameAndTypeInfo)
     * @param innermostClass the innermost class that surrounds this class.
     */
    public EnclosingMethod(ClassInfo innermostClass) {
        this(innermostClass, null);
    }

    /**
     * Returns information about the innermost class.
     *
     * @return information about the innermost class
     */
    public ClassInfo getInnermostClass() {
        return this.innermostClass;
    }

    /**
     * Returns the name and type info for the enclosing method. This method may
     * return null signifying there is no enclosing method.
     *
     * @return null if there is no enclosing method, the name and type of the
     * enclosing method otherwise
     */
    public NameAndTypeInfo<MethodDescriptor> getEnclosingMethod() {
        return this.enclosingMethod;
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2 + 2);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws ByteCodeException, IOException {
        super.write(out, cpool);
        innermostClass.writeIndex(out, cpool);
        if (enclosingMethod != null) {
            enclosingMethod.writeIndex(out, cpool);
        } else {
            U2.ZERO.write(out, cpool);
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        innermostClass.add(cpool);
        if (enclosingMethod != null) {
            enclosingMethod.add(cpool);
        }
    }
    @SuppressWarnings("unchecked")
    public static EnclosingMethod read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
        ClassInfo innermostClass = ccpt(cpool.get(in.readUnsignedShort()), ClassInfo.class, cpool);
        int nindex = in.readUnsignedShort();
        NameAndTypeInfo<MethodDescriptor> enclosingMethod = null;
        if (nindex != 0) {
            enclosingMethod = ccpt(cpool.get(nindex), NameAndTypeInfo.class, cpool);
            if (!enclosingMethod.isMethodDescriptor()) {
                throw new ByteCodeException("EnclosingMethod: Expected a method descriptor but found a field descriptor.");
            }
        }
        return new EnclosingMethod(innermostClass, enclosingMethod, name);
    }
}
