/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.attributes.innerclass;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.*;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.ConstantPool;
import teajay.bytecode.cpool.ConstantPoolAddable;
import teajay.bytecode.cpool.Utf8Info;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;
import static teajay.bytecode.util.Tools.ccpt;

/**
 * Implementation of the InnerClasses attribute as stated in "The JavaTM Virtual
 * Machine Specification Version 7". Revision: 27.07.2012
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class InnerClasses extends AttributeInfo {

    public static class InnerClassEntry implements ByteCodeWritable, ConstantPoolAddable {

        public final ClassInfo innerClass;
        public final ClassInfo outerClass;
        public final Utf8Info innerName;
        public final Collection<InnerClassAccessFlag> accessFlags;

        private InnerClassEntry(ClassInfo innerClass, ClassInfo outerClass, Utf8Info innerName, Collection<InnerClassAccessFlag> accessFlags) {
            this.innerClass = innerClass;
            this.outerClass = outerClass;
            this.innerName = innerName;
            this.accessFlags = Collections.unmodifiableCollection(VisibilityAccessFlags.handlePackagePrivate(accessFlags, InnerClassAccessFlag.class));
        }

        @Override
        public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
            innerClass.writeIndex(out, cpool);
            if (outerClass != null) {
                outerClass.writeIndex(out, cpool);
            } else {
                U2.ZERO.write(out, cpool);
            }
            if (innerName != null) {
                innerName.writeIndex(out, cpool);
            } else {
                U2.ZERO.write(out, cpool);
            }
            AccessFlagsIO.write(out, accessFlags);
        }

        @Override
        public void add(ConstantPool cpool) throws ByteCodeException {
            innerClass.add(cpool);
            if (outerClass != null) {
                outerClass.add(cpool);
            }
            if (innerName != null) {
                innerName.add(cpool);
            }
        }

        @Override
        public String toString() {
            return "inner: " + innerClass.getClassName() + ", outer: " + outerClass.getClassName()
                    + "\ninnerName: " + innerName.getValue() + ", access_flags: " + accessFlags;
        }

        public static InnerClassEntry read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
            ClassInfo innerClass = ccpt(cpool.get(in.readUnsignedShort()), ClassInfo.class, cpool);
            int outerIndex = in.readUnsignedShort();
            ClassInfo outerClass;
            if (outerIndex == 0) {
                outerClass = null;
            } else {
                outerClass = ccpt(cpool.get(outerIndex), ClassInfo.class, cpool);
            }
            int innerNameIndex = in.readUnsignedShort();
            Utf8Info innerName;
            if (innerNameIndex == 0) {
                innerName = null;
            } else {
                innerName = ccpt(cpool.get(innerNameIndex), Utf8Info.class, cpool);
            }
            Collection<InnerClassAccessFlag> flags = AccessFlagsIO.parseInnerClassAccessFlags(in.readUnsignedShort());
            return new InnerClassEntry(innerClass, outerClass, innerName, flags);
        }

        public ClassInfo getInnerClass() {
            return innerClass;
        }

        public ClassInfo getOuterClass() {
            return outerClass;
        }

        public Collection<InnerClassAccessFlag> getAccessFlags() {
            return accessFlags;
        }

        public Utf8Info getInnerName() {
            return innerName;
        }
    }
    public static final String NAME = "InnerClasses";
    private final List<InnerClassEntry> innerClasses = new ArrayList<>();

    /**
     * Creates a new InnerClasses attribute.
     */
    public InnerClasses() {
        this(new Utf8Info(NAME));
    }

    private InnerClasses(Utf8Info name) {
        super(name);
        checkName(NAME, name);
    }

    @Override
    public U4 getLength(ConstantPool cpool) throws ByteCodeException {
        return new U4(2 + innerClasses.size() * 8);
    }

    /**
     * Adds a new inner class to this attribute. <tt>innerName</tt> may be null
     * signifying that the class is anonymous. <tt>outerClass</tt> may be null
     * signifying that there is no outer class.
     *
     * @param innerClass the inner class's fully qualified name
     * @param outerClass the outer class's fully qualified name
     * @param innerName the name of the class as it is stated in the code
     * @param flags access flags
     */
    public void addInnerClass(ClassInfo innerClass, ClassInfo outerClass, Utf8Info innerName, InnerClassAccessFlag... flags) throws ByteCodeException {
        if (innerName == null) {
            if (outerClass != null) {
                throw new ByteCodeException("As of Java 7 it is not allowed to have the inner name index to be zero without the outer class index being zero aswell.");
            }
        }
        addInnerClass(new InnerClassEntry(innerClass, outerClass, innerName, Arrays.asList(flags)));
    }

    /**
     * @see #addInnerClass(teajay.bytecode.cpool.ClassInfo,
     * teajay.bytecode.cpool.ClassInfo, teajay.bytecode.cpool.Utf8Info,
     * teajay.bytecode.InnerClassAccessFlag[])
     * @param innerClass
     * @param outerClass
     * @param innerName
     * @param flags
     * @throws UTFOutOfBoundsException if the length of <tt>innerName</tt>
     * exceeds a 16 bit unsigned short
     */
    public void addInnerClass(ClassInfo innerClass, ClassInfo outerClass, String innerName, InnerClassAccessFlag... flags) throws UTFOutOfBoundsException, ByteCodeException {
        addInnerClass(innerClass, outerClass, innerName == null ? null : new Utf8Info(innerName), flags);
    }
    
    public InnerClassEntry findByInnerName(String innerName) {
        for (InnerClassEntry entry : innerClasses) {
            String tmp = entry.getInnerName().getValue();
            if (tmp != null && tmp.equals(innerName)) {
                return entry;
            }
        }
        return null;
    }
    
    public InnerClassEntry findByName(ClassInfo name) {
        for (InnerClassEntry entry : innerClasses) {
            ClassInfo tmp = entry.getInnerClass();
            if (tmp != null && tmp.equals(name)) {
                return entry;
            }
        }
        return null;
    }

    private void addInnerClass(InnerClassEntry entry) {
        innerClasses.add(entry);
    }

    public List<InnerClassEntry> getInnerClasses() {
        return Collections.unmodifiableList(innerClasses);
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws ByteCodeException, IOException {
        super.write(out, cpool);
        new U2(innerClasses.size()).write(out, cpool);
        for (InnerClassEntry cl : innerClasses) {
            cl.write(out, cpool);
        }
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        super.add(cpool);
        for (InnerClassEntry cl : innerClasses) {
            cl.add(cpool);
        }
    }

    public static InnerClasses read(DataInputStream in, ConstantPool cpool, int major, int minor, int length, Utf8Info name) throws ByteCodeException, IOException {
        int numEntries = in.readUnsignedShort();
        InnerClasses inner = new InnerClasses(name);
        for (int i = 0; i < numEntries; i++) {
            inner.addInnerClass(InnerClassEntry.read(in, cpool, major, minor));
        }
        return inner;
    }
}
