/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import java.io.*;
import java.util.*;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.cpool.MethodDescriptor.MParser;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.exceptions.UTFOutOfBoundsException;
import teajay.bytecode.methods.MethodKind;
import static teajay.bytecode.util.Tools.ccpt;
import teajay.bytecode.util.UnmodifiableIterator;

/**
 * Implementation of a ConstantPool as stated in "The JavaTM Virtual Machine
 * Specification Version 7". Revision: 27.07.2012 The constant pool is limited
 * to 65535 entries. Note: <tt>long</tt>s and
 * <tt>double</tt>s take up 2 slots.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public final class ConstantPool implements ByteCodeWritable, Iterable<CPoolEntry> {

    private static final DummyEntry dummy = new DummyEntry();

    private final SortedMap<Integer, CPoolEntry> entries;
    /**
     * Using a map here to speed up the contains query.
     *
     * @see #add(CPoolEntry).
     */
    private final Map<CPoolEntry, Integer> values;
    private int size = 1;

    public ConstantPool() {
        this.entries = new TreeMap<>();
        this.values = new HashMap<>();
    }

    private ConstantPool(int size) throws ByteCodeException {
        if (size > 0 && size <= 0xFFFF) {
            this.entries = new TreeMap<>();
            this.values = new HashMap<>((int) (size * 1.3), 0.75f);
            this.size = size;
        } else {
            throw new ByteCodeException("invalid constant pool size: " + size);
        }
    }

    /**
     * This method purges all entries from this pool.
     */
    public void purge() {
        entries.clear();
        values.clear();
    }

    private void set(int index, CPoolEntry entry) {
        if (index >= size) {
            throw new IndexOutOfBoundsException("constant pool index out of bounds: " + index);
        }
        entries.put(index, entry);
        values.put(entry, index);
    }

    public U2 getSize() throws ByteCodeException {
        return new U2(size);
    }

    public int size() {
        return size;
    }

    void add(CPoolEntry entry) throws ByteCodeException {
        boolean isDoubleOrLong;
        switch (entry.getTag()) {
            case Long:
            case Double:
                isDoubleOrLong = true;
                break;
            default:
                isDoubleOrLong = false;
        }
        if (size < 0xFFFE || (!isDoubleOrLong && size < 0xFFFF)) {
            Integer tmp = values.get(entry);
            if (tmp == null) { //not yet added
                //Add a new entry
                entries.put(size, entry);
                values.put(entry, size);

                if (isDoubleOrLong) {
                    entries.put(size + 1, dummy);
                    size = size + 2;
                } else {
                    size = size + 1;
                }
            }
        } else {
            throw new ByteCodeException("constant pool: not enough space to store the entry " + entry);
        }
    }

    public U2Index find(CPoolEntry entry) {
        Integer tmp = values.get(entry);
        if (tmp != null) {
            return new U2Index(tmp);
        }
        return null;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        getSize().write(out, cpool);
        for (CPoolEntry entry : entries.values()) {
            entry.write(out, cpool);
        }
    }

    @Override
    public Iterator<CPoolEntry> iterator() {
        return new UnmodifiableIterator<>(entries.values().iterator());
    }

    public CPoolEntry get(int index) throws ByteCodeException {
        CPoolEntry entry = entries.get(index);
        if (entry != null) {
            return entry;
        } else {
            throw new ByteCodeException("constant pool index out of bounds: " + index);
        }
    }

    public CPoolEntry get(U2Index index) throws ByteCodeException {
        return get(index.getValue());
    }

    public CPoolEntry get(U1Index index) throws ByteCodeException {
        return get(index.getValue());
    }

    @SuppressWarnings("unchecked")
    public static ConstantPool read(DataInputStream in, int major, int minor) throws IOException, ByteCodeException {
        try {
            MParser mparser = new MParser();
            int length = in.readUnsignedShort();
            DataInputStream[] pre = new DataInputStream[length];
            ConstantPool cpool = new ConstantPool(length);
            //Read the whole constant pool into memory for further processing
            for (int i = 1; i < length; i++) {
                CPoolTag tag = CPoolTag.read(in);

                switch (tag) {
                    case Class:
                    case String:
                    case MethodType:
                        pre[i] = read(in, 2, tag, -1);
                        break;
                    case FieldRef:
                    case MethodRef:
                    case InterfaceMethodRef:
                    case NameAndType:
                    case InvokeDynamic:
                        pre[i] = read(in, 4, tag, -1);
                        break;
                    case Utf8:
                        int strlen = in.readUnsignedShort();
                        pre[i] = read(in, 0, tag, strlen);
                        break;
                    case Integer:
                    case Float:
                        pre[i] = read(in, 4, tag, -1);
                        break;
                    case Long:
                    case Double:
                        pre[i] = read(in, 8, tag, -1);
                        pre[i + 1] = pre[i];
                        i++;
                        break;
                    case MethodHandle:
                        pre[i] = read(in, 3, tag, -1);
                        break;
                    default:
                        throw new ByteCodeException("unknown constant pool tag");
                }
            }

            for (int i = 1; i < length; i++) {
                DataInputStream bin = pre[i];
                CPoolTag tag = CPoolTag.read(bin);

                switch (tag) {
                    case Utf8:
                        /*
                         * Utf8Info.parse(String) automatically generates
                         * FieldDescriptor or MethodDescriptor instances from
                         * strings that satisfy the syntax. Otherwise it
                         * generates plain Utf8Info instances.
                         */
                        String str = bin.readUTF();
                        Utf8Info info = Utf8Info.parse(str, mparser);
                        if (!str.equals(info.getValue())) {
                            throw new Error("There is an internal error in Utf8Info.parse(String). Please report this bug to teajay@outx.de supplying this error message and the stack trace."
                                    + "\n Original: " + str
                                    + "\n Parsing result: " + info.getValue());
                        }
                        cpool.set(i, info);
                        break;
                    case Integer:
                        cpool.set(i, new IntegerInfo(bin.readInt()));
                        break;
                    case Float:
                        cpool.set(i, new FloatInfo(bin.readFloat()));
                        break;
                    case Long:
                        cpool.set(i, new LongInfo(bin.readLong()));
                        /*
                         * longs and doubles take up two constant pool addresses
                         */
                        cpool.set(++i, dummy);
                        break;
                    case Double:
                        cpool.set(i, new DoubleInfo(bin.readDouble()));
                        cpool.set(++i, dummy);
                        break;
                }
                pre[i].reset();
            }

            for (int i = 1; i < length; i++) {
                DataInputStream bin = pre[i];
                CPoolTag tag = CPoolTag.read(bin);

                switch (tag) {
                    case Class: {
                        Utf8Info name = ccpt(cpool.get(bin.readUnsignedShort()), Utf8Info.class, cpool);
                        cpool.set(i, new ClassInfo(name));
                        break;
                    }
                    case MethodType: {
                        MethodDescriptor descr = ccpt(cpool.get(bin.readUnsignedShort()), MethodDescriptor.class, cpool);
                        cpool.set(i, new MethodTypeInfo(descr));
                        break;
                    }
                    case NameAndType: {
                        Utf8Info name = ccpt(cpool.get(bin.readUnsignedShort()), Utf8Info.class, cpool);
                        Utf8Info descr = ccpt(cpool.get(bin.readUnsignedShort()), Utf8Info.class, cpool);
                        if (descr instanceof MethodDescriptor) {
                            cpool.set(i, new NameAndTypeInfo<>(name, (MethodDescriptor) descr));
                        } else if (descr instanceof FieldDescriptor) {
                            cpool.set(i, new NameAndTypeInfo<>(name, (FieldDescriptor) descr));
                        } else {
                            throw new ByteCodeException("invalid constant pool format: " + descr.getValue() + " not a descriptor");
                        }
                        break;
                    }
                    case String: {
                        Utf8Info value = ccpt(cpool.get(bin.readUnsignedShort()), Utf8Info.class, cpool);
                        cpool.set(i, new StringInfo(value));
                    }
                }
                pre[i].reset();
            }

            for (int i = 1; i < length; i++) {
                DataInputStream bin = pre[i];
                CPoolTag tag = CPoolTag.read(bin);

                switch (tag) {
                    case FieldRef: {
                        ClassInfo cinfo = ccpt(cpool.get(bin.readUnsignedShort()), ClassInfo.class, cpool);
                        NameAndTypeInfo<FieldDescriptor> ntype = ccpt(cpool.get(bin.readUnsignedShort()), NameAndTypeInfo.class, cpool);
                        if (ntype.isFieldDescriptor()) {
                            cpool.set(i, new FieldRefInfo(cinfo, ntype));
                        } else {
                            throw new ByteCodeException("invalid constant pool format");
                        }
                        break;
                    }
                    case MethodRef: {
                        ClassInfo cinfo = ccpt(cpool.get(bin.readUnsignedShort()), ClassInfo.class, cpool);
                        NameAndTypeInfo<MethodDescriptor> ntype = ccpt(cpool.get(bin.readUnsignedShort()), NameAndTypeInfo.class, cpool);
                        if (ntype.isMethodDescriptor()) {
                            cpool.set(i, new MethodRefInfo(cinfo, ntype));
                        } else {
                            throw new ByteCodeException("invalid constant pool format");
                        }
                        break;
                    }
                    case InterfaceMethodRef: {
                        ClassInfo cinfo = ccpt(cpool.get(bin.readUnsignedShort()), ClassInfo.class, cpool);
                        NameAndTypeInfo<MethodDescriptor> ntype = ccpt(cpool.get(bin.readUnsignedShort()), NameAndTypeInfo.class, cpool);
                        if (ntype.isMethodDescriptor()) {
                            cpool.set(i, new InterfaceMethodRefInfo(cinfo, ntype));
                        } else {
                            throw new ByteCodeException("invalid constant pool format");
                        }
                        break;
                    }
                }
                pre[i].reset();
            }

            for (int i = 1; i < length; i++) {
                DataInputStream bin = pre[i];
                CPoolTag tag = CPoolTag.read(bin);

                switch (tag) {
                    case MethodHandle: {
                        MethodKind kind = MethodKind.read(bin);
                        CPoolRef<?> ref = ccpt(cpool.get(bin.readUnsignedShort()), CPoolRef.class, cpool);
                        cpool.set(i, new MethodHandleInfo<>(kind, ref));
                        break;
                    }
                    case InvokeDynamic: {
                        U2 bootstrapIndex = new U2(bin.readUnsignedShort());
                        NameAndTypeInfo<MethodDescriptor> ntype = ccpt(cpool.get(bin.readUnsignedShort()), NameAndTypeInfo.class, cpool);
                        if (ntype.isMethodDescriptor()) {
                            cpool.set(i, new InvokeDynamicInfo(bootstrapIndex, ntype));
                        } else {
                            throw new ByteCodeException("invalid constant pool format");
                        }
                    }
                }
                pre[i].reset();
            }

            for (int i = 1; i < length; i++) {
                pre[i].close();
            }
            return cpool;
        } catch (IndexOutOfBoundsException | ClassCastException | ByteCodeException | UTFOutOfBoundsException ex) {
            throw new ByteCodeException("invalid constant pool format: " + ex.getMessage());
        }
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder();
        b.append("Constant pool size: ").append(size).append("\n");
        for (int i = 1; i < size; i++) {
            CPoolEntry entry = entries.get(i);
            try {
                CPoolIndex index = entry.getCIndex(this);
                b.append(index.getAddress(this)).append(": ").append(entry).append("\n");
            } catch (NullPointerException | ByteCodeException ex) {
                b.append("N/A: ").append(entry);
            }
        }
        return b.toString();
    }

    private static DataInputStream read(DataInputStream in, int len, CPoolTag tag, int stringLen) throws IOException {
        if (len < 0) {
            throw new IllegalArgumentException("len must not be negative");
        }
        byte[] buf = new byte[len + 1 + (stringLen < 0 ? 0 : stringLen + 2)];
        //rebuild what has already been read
        buf[0] = (byte) tag.intValue;
        int c = 1;
        if (stringLen >= 0) {
            buf[1] = (byte) ((stringLen >>> 8) & 0xFF);
            buf[2] = (byte) (stringLen & 0xFF);
            c += 2;
            len += stringLen;
        }
        len += c;
        int b = 0;
        while (len - c > 0 && (b = in.read(buf, c, len - c)) != -1) {
            c += b;
        }
        if (len != c) {
            throw new EOFException("unexpected end of file");
        }
        return new DataInputStream(new ByteArrayInputStream(buf));
    }
}
