/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.cpool;

import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Objects;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodKind;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 * @param <T>
 */
public final class MethodHandleInfo<T extends CPoolRef<? extends Utf8Info>> extends CPoolEntry {

    private final MethodKind referenceKind;
    private final T reference;

    public MethodHandleInfo(MethodKind referenceKind, T reference) throws IllegalArgumentException {
        super(CPoolTag.MethodHandle);
        this.referenceKind = referenceKind;
        this.reference = reference;
        switch (referenceKind) {
            case REF_getField:
            case REF_getStatic:
            case REF_putField:
            case REF_putStatic:
                if (!(reference instanceof FieldRefInfo)) {
                    throw new IllegalArgumentException("Invalid reference for " + referenceKind);
                }
                break;
            case REF_invokeVirtual:
            case REF_invokeStatic:
            case REF_invokeSpecial:
            case REF_newInvokeSpecial:
                if (!(reference instanceof MethodRefInfo)) {
                    throw new IllegalArgumentException("Invalid reference for " + referenceKind);
                }
                break;
            case REF_invokeInterface:
                if (!(reference instanceof InterfaceMethodRefInfo)) {
                    throw new IllegalArgumentException("Invalid reference for " + referenceKind);
                }
                break;
        }

        switch (referenceKind) {
            case REF_invokeVirtual:
            case REF_invokeStatic:
            case REF_invokeSpecial:
            case REF_invokeInterface:
                CPoolRef<?> info = (CPoolRef<?>) reference;
                String name = info.getNameAndType().getName().getValue();
                if (name.equals("<init>") || name.equals("<cinit>")) {
                    throw new IllegalArgumentException("Invalid method name for " + referenceKind);
                }
                break;
            case REF_newInvokeSpecial:
                info = (CPoolRef<?>) reference;
                name = info.getNameAndType().getName().getValue();
                if (!name.equals("<init>")) {
                    throw new IllegalArgumentException("Invalid method name for " + referenceKind);
                }
                break;
        }
    }
    
    public MethodKind getReferenceKind() {
        return referenceKind;
    }
    
    public T getReference() {
        return reference;
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        super.write(out, cpool);
        referenceKind.write(out, cpool);
        reference.getCIndex(cpool).write(out, cpool);
    }
    
    @Override
    public boolean equals(Object obj) {
        if ((obj instanceof MethodHandleInfo) && super.equals(obj)) {
            MethodHandleInfo<?> other = (MethodHandleInfo<?>) obj;
            return other.getReference().equals(getReference()) && other.getReferenceKind().equals(getReferenceKind());
        } else {
            return false;
        }
    }

    @Override
    public int hashCode() {
        int hash = 41 * super.hashCode();
        hash = 97 * hash + (this.referenceKind != null ? this.referenceKind.hashCode() : 0);
        hash = 97 * hash + Objects.hashCode(this.reference);
        return hash;
    }
    
    @Override
    public void add(ConstantPool pool) throws ByteCodeException {
        super.add(pool);
        reference.add(pool);
    }
}
