/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.bytecode.fields;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.*;
import teajay.bytecode.AccessFlagsIO;
import teajay.bytecode.ByteCodeWritable;
import teajay.bytecode.U2;
import teajay.bytecode.VisibilityAccessFlags;
import teajay.bytecode.attributes.AttributeInfo;
import teajay.bytecode.attributes.AttributesReader;
import teajay.bytecode.cpool.*;
import teajay.bytecode.exceptions.ByteCodeException;
import static teajay.bytecode.util.Tools.ccpt;
import static teajay.bytecode.util.Tools.checkUnqualifiedName;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class FieldInfo implements ByteCodeWritable, ConstantPoolAddable {

    private final Collection<FieldAccessFlag> accessFlags;
    private final Utf8Info name;
    private final FieldDescriptor descriptor;
    private final Map<String, AttributeInfo> attributes;

    public FieldInfo(Utf8Info name, FieldDescriptor descriptor, FieldAccessFlag... accessFlags) throws ByteCodeException {
        this(name, descriptor, Arrays.asList(VisibilityAccessFlags.handlePackagePrivate(accessFlags, FieldAccessFlag.class)));
    }

    private FieldInfo(Utf8Info name, FieldDescriptor descriptor, Collection<FieldAccessFlag> accessFlags) throws ByteCodeException {
        this.accessFlags = Collections.unmodifiableCollection(accessFlags);
        this.name = name;
        /*
         * The Oracle JVM does not enforce field names to be unqualified names.
         */
        if (!checkUnqualifiedName(name)) {
            throw new ByteCodeException("FieldInfo: not a unqualified name " + name.getValue());
        }
        this.descriptor = descriptor;
        this.attributes = new HashMap<>();
    }

    public FieldInfo(String name, FieldDescriptor descriptor, FieldAccessFlag... accessFlags) throws ByteCodeException {
        this(new Utf8Info(name), descriptor, accessFlags);
    }

    public FieldRefInfo getFieldRefInfo(ClassInfo cl) {
        return new FieldRefInfo(cl, new NameAndTypeInfo<>(name, descriptor));
    }

    public void addAttribute(AttributeInfo info) {
        attributes.put(info.getName().getValue(), info);
    }

    public Collection<AttributeInfo> getAttributes() {
        return attributes.values();
    }

    public AttributeInfo getAttribute(String name) {
        return attributes.get(name);
    }

    public void removeAttribute(String name) {
        attributes.remove(name);
    }

    public Collection<FieldAccessFlag> getAccessFlags() {
        return accessFlags;
    }

    public Utf8Info getName() {
        return name;
    }

    public String getFieldName() {
        return name.getValue();
    }

    public FieldDescriptor getDescriptor() {
        return descriptor;
    }

    public boolean isStatic() {
        return accessFlags.contains(FieldAccessFlag.STATIC);
    }

    public boolean isVolatile() {
        return accessFlags.contains(FieldAccessFlag.VOLATILE);
    }

    public boolean isTransient() {
        return accessFlags.contains(FieldAccessFlag.TRANSIENT);
    }

    public boolean isEnum() {
        return accessFlags.contains(FieldAccessFlag.ENUM);
    }

    public boolean isPublic() {
        return accessFlags.contains(FieldAccessFlag.PUBLIC);
    }

    public boolean isPrivate() {
        return accessFlags.contains(FieldAccessFlag.PRIVATE);
    }

    public boolean isProtected() {
        return accessFlags.contains(FieldAccessFlag.PROTECTED);
    }

    public boolean isPackagePrivate() {
        return accessFlags.contains(FieldAccessFlag.PACKAGE_PRIVATE);
    }

    public boolean isFinal() {
        return accessFlags.contains(FieldAccessFlag.FINAL);
    }

    public boolean isSynthetic() {
        return accessFlags.contains(FieldAccessFlag.SYNTHETIC);
    }

    @Override
    public void add(ConstantPool cpool) throws ByteCodeException {
        name.add(cpool);
        descriptor.add(cpool);
        for (AttributeInfo info : attributes.values()) {
            info.add(cpool);
        }
    }

    @Override
    public void write(DataOutputStream out, ConstantPool cpool) throws IOException, ByteCodeException {
        AccessFlagsIO.write(out, accessFlags);
        name.getCIndex(cpool).write(out, cpool);
        descriptor.getCIndex(cpool).write(out, cpool);
        new U2(attributes.size()).write(out, cpool);
        for (AttributeInfo attr : attributes.values()) {
            attr.write(out, cpool);
        }
    }

    @Override
    public String toString() {
        return getClass().getSimpleName() + ": name=" + name.getValue() + ", descriptor=" + descriptor.getValue();
    }

    public static FieldInfo read(DataInputStream in, ConstantPool cpool, int major, int minor) throws ByteCodeException, IOException {
        Collection<FieldAccessFlag> flags = AccessFlagsIO.parseFieldAccessFlags(in.readUnsignedShort());
        Utf8Info name = ccpt(cpool.get(in.readUnsignedShort()), Utf8Info.class, cpool);
        FieldDescriptor descriptor = ccpt(cpool.get(in.readUnsignedShort()), FieldDescriptor.class, cpool);
        FieldInfo field = new FieldInfo(name, descriptor, flags);

        int attributesCount = in.readUnsignedShort();
        for (int i = 0; i < attributesCount; i++) {
            field.addAttribute(AttributesReader.read(in, cpool, major, minor));
        }

        return field;
    }
}
