/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package codegeneration.environment;

import abstractsyntaxtree.SyntaxNode;
import abstractsyntaxtree.TjTypeDefNode;
import datastructs.Pair;
import exceptions.CompilerException;
import java.lang.invoke.CallSite;
import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import symbol.SymbolFacade;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.bootstrap.BootstrapMethodEntry;
import teajay.bytecode.attributes.bootstrap.BootstrapMethods;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.InvokeDynamicInfo;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.cpool.MethodHandleInfo;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.cpool.NameAndTypeInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodInfo;
import teajay.bytecode.methods.MethodKind;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.Type;

/**
 *
 * @author Michael Burzan
 * Environment to manage information about methods, 
 * generics, derivations, types and fields
 */
public class Stage2Environment implements Comparable<Stage2Environment>{
    private final Stage1Environment env;
    private Type.ClassSigType signature;
    private LinkedList<HashSet<String>> genericEnv = new LinkedList<>();
    private TypeUtils typeUtils;
    private LinkedList<MethodInfo> methodList = new LinkedList<>();
    private BootstrapMethods bootstrapMethods;
    private MethodHandleInfo staticBootstrap;
    private TjTypeDefNode typeDefNode =null;

    public MethodHandleInfo getStaticBootstrapMethod() {
        return staticBootstrap;
    }

    public BootstrapMethods getBootstrapMethods() {
        return bootstrapMethods;
    }
    
    
    public Stage2Environment(Stage1Environment env){
        this.env = env;
        try {
            /*
             * Create bootstrap info to call the runtinme environment
             */
            MethodRefInfo ref = MethodRefInfo.make("teajay/runtime/Bootstraps", "st",
                    FieldDescriptor.fromClass(CallSite.class), FieldDescriptor.fromClass(MethodHandles.Lookup.class),
                    FieldDescriptor.getString(), FieldDescriptor.fromClass(MethodType.class), FieldDescriptor.fromClass(Class.class));
            staticBootstrap = new MethodHandleInfo(MethodKind.REF_invokeStatic, ref);
            this.bootstrapMethods = (BootstrapMethods) this.env.getClassFile().getAttribute(BootstrapMethods.NAME);
            if (this.bootstrapMethods == null) {
                this.bootstrapMethods = new BootstrapMethods();
                this.env.getClassFile().addAttribute(bootstrapMethods);
            }
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }
    /**
     * Create the invoke dynamic instruction to call static methods
     * @param typedefMethod static method info
     * @return invoke dynamic instruction
     * @throws CompilerException 
     */
    public InvokeDynamicInfo invokeDynamic(MethodInfo typedefMethod) throws CompilerException {
        try {
            BootstrapMethodEntry addBoostrapMethod = this.bootstrapMethods.addBootstrapMethod(this.staticBootstrap, typedefMethod.getDefiningClass());
            NameAndTypeInfo<MethodDescriptor> nInfo = new NameAndTypeInfo<>(typedefMethod.getName().getValue(), typedefMethod.getDescriptor());
            InvokeDynamicInfo dynamicInfo = new InvokeDynamicInfo(addBoostrapMethod, nInfo);
            return dynamicInfo;
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return null;
    }


    /**
     * get access to symbol facade
     * @return  current symbol facade
     */
    public SymbolFacade getSymbolFacade() {
        return this.env.getFacade();
    }
    /**
     * 
     * @return the current class file object 
     */
    public ClassFile getClassFile() {
        return env.getClassFile();
    }
    /**
     * save the class signature
     * @param signature signature of type
     */
    public void setClassSigType(Type.ClassSigType signature){
        this.signature = signature;
    }
    /**
     * create a new generic context
     * Example:
     * typedef stack<T> {
     *  static <X> void foo(X x); // new Frame
     * }
     */
    public void addGenericFrame(){
        genericEnv.addLast(new HashSet<String>());
    }
    /**
     * remove the last generic frame
     */
    public void dropGenericFrame(){
        genericEnv.removeLast();
    }
    /**
     * add new type variable to current generic frame.
     * Example:
     * typedef stack<T>{
     * }
     * add T to current frame
     * @param identifier name of the typevariable 
     */
    public void addTypeVariable(String identifier) {
        genericEnv.getLast().add(identifier);
    }
    /**
     * 
     * @param identifier name of the typevariable
     * @return true if identifier stored in a frame
     */
    public boolean isTypeVariable(String identifier){
        Iterator<HashSet<String>> descendingIterator = genericEnv.descendingIterator();
        while(descendingIterator.hasNext()){
            if(descendingIterator.next().contains(identifier)){
                return true;
            }
        }
        return false;
    }

    public void setTypeUtils(TypeUtils typeUtils) {
        this.typeUtils = typeUtils;
    }
    
    public TypeUtils getTypeUtils(){
        return this.typeUtils;
    }

    @Override
    public int compareTo(Stage2Environment o) {
        return this.env.compareTo(o.env);
    }

    public Pair<String,SyntaxNode> getAst() {
        return env.getAst();
    }

    public void pushMethodInfo(MethodInfo info) {
        this.methodList.add( info );
        
    }
    public MethodInfo popMethodInfo(){
        return this.methodList.removeFirst();
    }

    Stage1Environment getStage1Env() {
        return this.env;
    }

    public void addTypeDefNode(TjTypeDefNode node) {
        this.typeDefNode = node;
    }
    
    public TjTypeDefNode getTypeDefNodes(){
        return this.typeDefNode;
    }

    private LinkedList<HashSet<String>> tmpGenericEnv;
    /**
     * Save and overwrite the current generic environment with all frames
     * @param typeVariables  
     */
    public void setTypeVariables(LinkedList<HashSet<String>> typeVariables) {
        tmpGenericEnv = this.genericEnv;
        this.genericEnv = typeVariables;
    }
    /**
     * Restore the old generic environment
     */
    public void resetGenericEnv(){
        this.genericEnv = tmpGenericEnv;
    }
    
    
    
}
