package lexer;
import exceptions.LexerException;
import java.io.IOException;
import static lexer.UTF16.utf16Decode;

/**
 * This Reader maps unicode escape sequences to its respective characters or if
 * necessary to its code points.
 * @author Michael Kußmann <michi@outx.de>
 */
%%
%public
%final
%class UnicodeFilteredReader
%extends java.io.Reader
%yylexthrow LexerException
%int
%unicode
%apiprivate
%line
%column
%javanl
%ctorarg int initialBufferSize
%init{
    if (initialBufferSize < 0) {
        throw new IllegalArgumentException("initial buffer size must not be negative");
    }
    this.buffer = new StringBuilder(initialBufferSize);
%init}

%{
    private StringBuilder buffer;
    private Character utf16BlockTmp = null;
    private int minLexLength = -1;
    private int currentLexLength = 0;
    private String sourceFile = null;

    public UnicodeFilteredReader(java.io.Reader in) {
        this(in, ZZ_BUFFERSIZE + 16);
    }

    public UnicodeFilteredReader(java.io.InputStream in) {
        this(in, ZZ_BUFFERSIZE + 16);
    }

    public void setSourceFile(String filename) {
        this.sourceFile = filename;
    }

    @Override
    public int read(char[] cbuf, int off, int len) throws IOException {
        if (len < 0) {
            throw new IndexOutOfBoundsException("len must not be negative");
        } else if (len == 0) {
            return 0;
        }
        synchronized(this.lock) {
            if (buffer == null) {
                throw new IOException("stream closed");
            }

            this.minLexLength = len - buffer.length();
            this.currentLexLength = 0;
            if (this.minLexLength > 0) {
                yylex();
            }
            this.minLexLength = -1;
            this.currentLexLength = 0;

            int read;
            int buflen = buffer.length();
            if (buflen == 0) {
                read = -1;
            } else if (len <= buflen) {
                read = len;
                buffer.getChars(0, len, cbuf, off);
                buffer.delete(0, len);
            } else {
                read = buflen;
                buffer.getChars(0, buflen, cbuf, off);
                buffer.delete(0, buflen);
            }
            return read;
        }
    }

    @Override
    public void close() throws IOException {
        synchronized(this.lock) {
            yyclose();
            buffer = null;
        }
    }

    private void error(String msg) throws LexerException {
        if (sourceFile != null) {
            throw new LexerException(msg, sourceFile, yyline + 1, yycolumn + 1);
        } else {
            throw new LexerException(msg, yyline + 1, yycolumn + 1);
        }
    }

    private void error(String msg, int line, int column) throws LexerException {
        if (sourceFile != null) {
            throw new LexerException(msg, sourceFile, line + 1, column + 1);
        } else {
            throw new LexerException(msg, line + 1, column + 1);
        }
    }
%}

HexDigit = [0-9a-fA-F]

%state UNICODE_ESCAPE

%%
<YYINITIAL> {
    (\\\\){1,16}
                {   utf16BlockTmp = null;
                    int yylen = zzMarkedPos - zzStartRead;
                    buffer.append(zzBuffer, zzStartRead, yylen);
                    currentLexLength += yylen;
                    if (currentLexLength >= minLexLength) {
                        return 0;
                    }
                }

    \\u+        { yybegin(UNICODE_ESCAPE); }

    \\          {   utf16BlockTmp = null;
                    buffer.append('\\');
                    currentLexLength++;
                    if (currentLexLength >= minLexLength) {
                        return 0;
                    }
                }

    [^\\]{1,256}
                {   utf16BlockTmp = null;
                    int yylen = zzMarkedPos - zzStartRead;
                    buffer.append(zzBuffer, zzStartRead, yylen);
                    currentLexLength += yylen;
                    if (currentLexLength >= minLexLength) {
                        return 0;
                    }
                }
}
<UNICODE_ESCAPE> {
    {HexDigit}{4}   {   yybegin(YYINITIAL); 
                        char code = (char) Integer.parseInt(yytext(), 16);
                        Integer utf32;
                        if (utf16BlockTmp != null && (utf32 = utf16Decode(utf16BlockTmp, code)) != null) {
                            buffer.deleteCharAt(buffer.length() - 1);
                            buffer.appendCodePoint(utf32);
                            utf16BlockTmp = null;
                        } else {
                            utf16BlockTmp = code;
                            buffer.append(code);
                        }
                        currentLexLength++;
                        if (utf16BlockTmp == null && currentLexLength >= minLexLength) {
                            return 0;
                        }
                    }

    [^]             { yybegin(YYINITIAL); error("illegal unicode escape"); }
}