/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package minicompiler;

import abstractsyntaxtree.SyntaxNode;
import codegeneration.environment.Stage0Environment;
import codegeneration.environment.Stage1Environment;
import codegeneration.environment.Stage2Environment;
import codegeneration.environment.Stage3Environment;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import lexer.LookAheadLexer;
import parser.Parser;
import stage0.Stage0Visitor;
import stage1.Stage1Visitor;
import stage2.Stage2CompilationUnitVisitor;
import stage3.Stage3CompilationUnitVisitor;
import stage3.utils.InformationCenter;
import symbol.DerivationTree;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.SourceFile;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.Type;
import teajay.generics.writer.SignatureEncoder;
import teajay.methods.MethodFinder;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;
import teajay.util.io.Input;
import teajay.util.tjattribute.TJBridge;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael Burzan
 */
public class Compiler {

    private List<Pair<String, SyntaxNode>> astList = new ArrayList<>();
    private LinkedList<Stage0Environment> stage0EnvList = new LinkedList<>();
    private LinkedList<Stage1Environment> stage1EnvList = new LinkedList<>();
    private LinkedList<Stage2Environment> stage2EnvList = new LinkedList<>();
    private HashSet<WorkingNode> workingProvider = new HashSet<>();

    public boolean compile() throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        this.compileFilesFromUserInput();


        this.compileStage0();//Imports and Names
        this.compileStage1();// Types,
        this.compileStage2();// Derivations,implementations, fields, methods,generic context
        this.compileStage3();//rest
        return true;
    }

    private void compileFilesFromUserInput() throws FileNotFoundException, CompilerException, IOException {
        CompileManager manager = CompileManager.instance();
        while (manager.hasFiles()) {
            File file = manager.popFile();
            //System.err.println("Compiling file: " + file.getAbsolutePath());
            try (LookAheadLexer lexer = new LookAheadLexer(file, CompileManager.instance().encoding())) {
                CompileManager.instance().setCurrentFilename(file.toString());
                Parser parser = new Parser(lexer);
                SyntaxNode root = parser.parse();
                PlainTypenameVisitor plainTypenameVisitor = new PlainTypenameVisitor();
                root.accept(plainTypenameVisitor);
                String qid = plainTypenameVisitor.getQID();
                if (qid != null) {
                    this.astList.add(new Pair<>(file.toString(), root));
                    SourceFileNameChecker.checkSourceFile(file.getName(), qid, root);
                    workingProvider.add(new WorkingNode(null, qid, null, true));
                } else {
                    System.err.println("[WARNING] file without entity declaration: " + file);
                }
            }
        }
    }

    private void compileFilesFromStage0(Set<WorkingNode> listOfWorkingQueues) throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        if (!listOfWorkingQueues.isEmpty()) {
            this.astList.clear();
            for (WorkingNode node : listOfWorkingQueues) {
                Collection<Input> inputs = node.getInputs();
                for (Input input : inputs) {
                    try (LookAheadLexer lexer = new LookAheadLexer(
                            new InputStreamReader(input.openStream(), CompileManager.instance().encoding()), input.toString())) {
                        CompileManager.instance().setCurrentFilename(input.toString());
                        Parser parser = new Parser(lexer);
                        this.astList.add(new Pair<>(input.toString(), parser.parse()));
                    }
                }
            }

            this.compileStage0();
        }
    }

    private void compileStage0(Pair<String, SyntaxNode> ast, HashSet<WorkingNode> listOfWorkingQueues) throws FileNotFoundException, CompilerException, IOException {
        Stage0Environment env = new Stage0Environment(ast);
        //env.setAst(ast);

        CompileManager.instance().setCurrentFilename(ast.value1);
        ast.value2.accept(new Stage0Visitor(env));
        addWorkingList(listOfWorkingQueues, env.getWorkingQueueForCompilationUnit());
        //listOfWorkingQueues.addAll(env.getWorkingQueueForCompilationUnit());
        this.stage0EnvList.add(env);

    }

    private void compileStage0() throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        HashSet<WorkingNode> listOfWorkingQueues = new HashSet<>();
        for (Pair<String, SyntaxNode> ast : astList) {
            compileStage0(ast, listOfWorkingQueues);
        }
        compileFilesFromStage0(listOfWorkingQueues);
        for (WorkingNode node : listOfWorkingQueues) {
            node.setCompileFileToFacade();
        }

    }

    private void compileStage1() throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        HashSet<WorkingNode> listOfWorkingQueues = new HashSet<>();
        Collections.sort(this.stage0EnvList);
        for (Stage0Environment env : this.stage0EnvList) {
            Pair<String, SyntaxNode> ast = env.getAst();
            Stage1Environment stage1Env = new Stage1Environment(env);
            CompileManager.instance().setCurrentFilename(ast.value1);
            ast.value2.accept(new Stage1Visitor(stage1Env));
            this.stage1EnvList.add(stage1Env);
            addWorkingList(listOfWorkingQueues, stage1Env.getWorkingList());
            //listOfWorkingQueues.addAll(stage1Env.getWorkingList());
            env.getClassFile().addAttribute(new SourceFile(new File(ast.value1).getName()));
        }
        compileFilesFromStage1(listOfWorkingQueues);
    }

    private void compileStage2() throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        Collections.sort(stage1EnvList);
        List<Pair<ClassFile, SyntaxNode>> typeDefList = new LinkedList<>();
        insertSymbolToFacade();


        for (Stage1Environment env : this.stage1EnvList) {
            Pair<String, SyntaxNode> ast = env.getAst();
            Stage2Environment stage2Env = new Stage2Environment(env);
            CompileManager.instance().setCurrentFilename(ast.value1);

            ClassFile classFile = stage2Env.getClassFile();
            TeaJayAttribute.Type teaJayAttributeType = InformationCenter.getTeaJayAttributeType(classFile);
            if (teaJayAttributeType == TeaJayAttribute.Type.Typedef) {
                typeDefList.add(new Pair<>(classFile, ast.value2));
                //addInterfaceMethods(stage2Env, ast.value2);
            } else if (teaJayAttributeType == TeaJayAttribute.Type.Impl) {
                for (Pair<ClassFile, SyntaxNode> td : typeDefList) {

                    addBridgeMethods(td.value1, td.value2);
                }
                typeDefList.clear();
            }

            ast.value2.accept(new Stage2CompilationUnitVisitor(stage2Env));

            this.stage2EnvList.add(stage2Env);
        }
        for (Pair<ClassFile, SyntaxNode> td : typeDefList) {

            addBridgeMethods(td.value1, td.value2);
        }
        typeDefList.clear();
    }

    private static void addInterfaceMethods(Stage2Environment env, SyntaxNode node) throws CompilerException {
        if (true) {
            return;
        }
        try {
            ClassFile file = env.getClassFile();
            Set<MethodInfo> methods = new HashSet<>();
            for (ClassInfo iface : file.getInterfaces()) {

                ClassFileRef ref = ClassFinder.instance().find(iface.getClassName());
                if (ref == null) {
                    throw new CompilerException("Type " + iface.getClassName() + " not found", node.getSymbol());
                }
                for (List<MethodInfo> lst : ref.get().getMethods()) {
                    for (MethodInfo info : lst) {
                        methods.add(info);
                    }
                }

            }
            for (MethodInfo info : methods) {
                assert info.isAbstract();

                MethodInfo newInfo = info.makeFresh(file.getThis(), info.getAccessFlags());
                if (!file.hasMethod(newInfo.getName().getValue(), newInfo.getDescriptor())) {
                    file.addMethod(newInfo);
                }
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    private void compileStage3() throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        Collections.sort(stage2EnvList);
        for (Stage2Environment env : stage2EnvList) {
            Pair<String, SyntaxNode> ast = env.getAst();
            Stage3Environment stage3Env = new Stage3Environment(env);
            CompileManager.instance().setCurrentFilename(ast.value1);

            ast.value2.accept(new Stage3CompilationUnitVisitor(stage3Env));
            CompileManager.instance().writeClassFile(stage3Env.getClassFile(), stage3Env);
        }
    }

    private void compileFilesFromStage1(HashSet<WorkingNode> listOfWorkingQueues) throws FileNotFoundException, CompilerException, IOException, ByteCodeException, ResolveException {
        if (!stage0EnvList.isEmpty()) {
            stage0EnvList.clear();
            compileFilesFromStage0(listOfWorkingQueues);
            for (WorkingNode node : listOfWorkingQueues) {
                node.setCompileFileToFacade();

            }
            compileStage1();
        }
    }

    private void addBridgeMethods(ClassFile file, SyntaxNode ast) throws ByteCodeException, IOException, ResolveException, CompilerException {
        Collection<List<MethodInfo>> tmpMethods = file.getMethods();
        ClassInfo superTypeDefInfo = file.getSuper();
        if (superTypeDefInfo == null || superTypeDefInfo.getClassName().equals("java/lang/Object")) {
            return;
        }
        Collection<List<MethodInfo>> methods = new ArrayList<>();
        methods.addAll(tmpMethods);

        for (List<MethodInfo> mTemp : methods) {
            List<MethodInfo> lInfo = new ArrayList<>();
            lInfo.addAll(mTemp);
            for (MethodInfo method : lInfo) {
                if (method.getName().getValue().equals("<init>") || method.isPrivate() || method.isStatic()) {
                    continue;
                }
                Pair<MethodInfo, MethodFinder.DistanceTailPair> findMethod = MethodFinder.findMethod(method.getName().getValue(),
                        method.getDescriptor().getParameters(),
                        superTypeDefInfo.getClassName(), file.getThis().getClassName(),
                        ast, MethodFinder.QueryType.NON_STATIC_EXACT);
                for (ClassInfo iface : file.getInterfaces()) {
                    if (findMethod == null) {
                        findMethod = MethodFinder.findMethod(method.getName().getValue(),
                                method.getDescriptor().getParameters(),
                                iface.getClassName(), file.getThis().getClassName(),
                                ast, MethodFinder.QueryType.NON_STATIC_EXACT);
                    } else {
                        break;
                    }
                }
                if (findMethod != null) {
                    FieldDescriptor baseReturnType = findMethod.value1.getDescriptor().getReturnType();
                    FieldDescriptor methodReturnType = method.getDescriptor().getReturnType();
                    ClassFileRef ref1 = ClassFinder.instance().find(baseReturnType.getClassInfo().getClassName());
                    ClassFileRef ref2 = ClassFinder.instance().find(methodReturnType.getClassInfo().getClassName());
                    if (DerivationTree.instance().isAssignableAndDifferent(ref1, ref2)) {

                        Collection<MethodAccessFlag> tmp = findMethod.value1.getAccessFlags();
                        Collection<MethodAccessFlag> accessFlags = new LinkedList<>();
                        accessFlags.addAll(tmp);
                        accessFlags.add(MethodAccessFlag.BRIDGE);
                        MethodInfo bridgeMethod = new MethodInfo(findMethod.value1.getName().getValue(),
                                findMethod.value1.getDescriptor(),
                                file.getThis(), accessFlags);
                        bridgeMethod.addAttribute(new TJBridge(SignatureEncoder.encode(TypeUtils.getMethodSignature(method))));

                        file.addMethod(bridgeMethod);

                    } else if (!DerivationTree.instance().isAssignable(ref1, ref2)) {
                        throw new CompilerException(method.getName().getValue()
                                + " in " + file.getThis().getClassName() + "can not override "
                                + method.getName().getValue() + " in " + superTypeDefInfo.getClassName()
                                + ": return types " + ref1.get().getThis().getClassName()
                                + " and " + ref2.get().getThis().getClassName()
                                + " are not assignable", ast.getSymbol());
                    }

                }


            }
        }



    }

    private void addWorkingList(HashSet<WorkingNode> listOfWorkingQueues, HashSet<WorkingNode> workingList) {
        for (WorkingNode node : workingList) {
            if (!this.workingProvider.contains(node)) {
                listOfWorkingQueues.add(node);
                workingProvider.add(node);
            }
        }
    }

    private void insertSymbolToFacade() throws CompilerException, ByteCodeException, ResolveException, IOException {
        List<String> classFileNames = new ArrayList<>();
        for (Stage1Environment env : this.stage1EnvList) {
            ClassFile classFile = env.getEnv().getClassFile();
            if (ClassFinder.instance().find(classFile.getThis().getClassName()) == null) {
                throw new Error("Fatal error: ClassFile object with name " + classFile.getThis().getClassName() + " not found");
            }

            if (env.getFacade().getFullQualifiedId(classFile.getThis().getSimpleName()) == null) {
                classFileNames.add(classFile.getThis().getClassName());
            }
        }
        for (String s : classFileNames) {
            for (Stage1Environment env : this.stage1EnvList) {
                env.getFacade().addNameFromPackage(s);
            }
        }
    }
}

/*
 * typedef [abstract] AbstractBank {
 * 
 *      AbstractBank(Number zins);
 *      void setZins(Number zins);
 *      Number getZins();
 *      
 * }
 * 
 * typeimpl AbstractBankImpl implements AbstractBank {
 * 
 *      
 *      AbstractBank(Number zins){
 *          this.zins = zins;
 *      }
 * 
 *      Number getZins(){
 *          return zins;    
 *      }
 * 
 * 
 * }
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 * 
 */
