/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage2.utils;

import abstractsyntaxtree.ModifierNode;
import static abstractsyntaxtree.ModifierNode.Modifier.ABSTRACT;
import static abstractsyntaxtree.ModifierNode.Modifier.FINAL;
import static abstractsyntaxtree.ModifierNode.Modifier.NATIVE;
import static abstractsyntaxtree.ModifierNode.Modifier.PRIVATE;
import static abstractsyntaxtree.ModifierNode.Modifier.PROTECTED;
import static abstractsyntaxtree.ModifierNode.Modifier.PUBLIC;
import static abstractsyntaxtree.ModifierNode.Modifier.STATIC;
import static abstractsyntaxtree.ModifierNode.Modifier.STRICTFP;
import static abstractsyntaxtree.ModifierNode.Modifier.SYNCHRONIZED;
import static abstractsyntaxtree.ModifierNode.Modifier.SYNTHETIC;
import static abstractsyntaxtree.ModifierNode.Modifier.TRANSIENT;
import static abstractsyntaxtree.ModifierNode.Modifier.VOLATILE;
import abstractsyntaxtree.SyntaxNode;
import codegeneration.environment.Stage2Environment;
import codegeneration.javavisitor.QualifiedIdentifierListVisitor;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import stage2.Stage2FormalParametersVisitor;
import symbol.DerivationTree;
import teajay.bytecode.ClassFile;
import teajay.bytecode.attributes.Exceptions;
import teajay.bytecode.attributes.signature.Signature;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.MethodDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.fields.FieldAccessFlag;
import teajay.bytecode.methods.MethodAccessFlag;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.SignatureParser;
import teajay.generics.parser.Type;
import teajay.generics.writer.SignatureEncoder;
import teajay.methods.MethodFinder;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
public class MethodHeadBuilder {

    private static Stage2Environment xxxEnv;

    /**
     * this method generate typeimpl method heads
     *
     * @param node
     * @param stage2Env
     * @param returnType the return type from the method
     * @param methodName the name from the metho
     * @param methodHead parameters and throws list
     * @param flagList modifier
     * @throws CompilerException
     */
    public static void buildMethodHeadTypeImpl(SyntaxNode node, Stage2Environment stage2Env,
            Type returnType, String methodName,
            Pair<List<Pair<Type, String>>, List<String>> methodHead,
            Collection<ModifierNode.Modifier> flagList) throws CompilerException {

        xxxEnv = stage2Env;
        MethodAccessFlag[] accessFlags;

        List<MethodAccessFlag> outList = new LinkedList<>();
        //check the access flags 
        Pair<Boolean, Boolean> methodAccess = MethodHeadBuilder.checkAccessFlagsForTypeImpl(node, outList, flagList);
        boolean isMethodPublic = methodAccess.value1;
        boolean isMethodPrivate = methodAccess.value2;
        accessFlags = outList.toArray(new MethodAccessFlag[0]);
        MethodInfo info = createMethodInfo(stage2Env, returnType, methodName, methodHead, accessFlags, node);
        if (!methodName.equals("<init>")) {
            //push method info to environment
            stage2Env.pushMethodInfo(info);
        } else {
            //build constructor if method <init>
            ConstructorBuilder.build(info, stage2Env);
        }
        //exist methoddefiniation in typeimpl
        checkExistMethodInClassFile(stage2Env.getClassFile(), methodName,
                info, node);
        if (isMethodPublic && !methodName.equals("<init>")) {
            //exist method in typedef
            checkIsMethodInTypeDef(stage2Env.getClassFile(), methodName, info, node);
        } else if (isMethodPublic) {
            try {
                List<FieldDescriptor> newParams = new LinkedList<>();
                newParams.addAll(info.getDescriptor().getParameters().subList(1, info.getDescriptor().getParameterCount()));
                MethodDescriptor descr = new MethodDescriptor(info.getDescriptor().getReturnType(), newParams);
                MethodInfo dummy = new MethodInfo(info.getName().getValue(), descr, info.getDefiningClass(), MethodAccessFlag.PUBLIC);
                if (info.getAttribute(Signature.NAME) != null) {
                    Type.MethodType parse = (Type.MethodType) SignatureParser.parse(((Signature) info.getAttribute(Signature.NAME)).getSignature());
                    parse.paramTypes.remove(0);
                    dummy.addAttribute(new Signature(SignatureEncoder.encode(parse)));
                }
                if (info.getAttribute(Exceptions.NAME) != null) {
                    dummy.addAttribute(info.getAttribute(Exceptions.NAME));
                }

                checkIsMethodInTypeDef(stage2Env.getClassFile(), methodName, dummy, node);
            } catch (ByteCodeException ex) {
                ex.printStackTrace(System.err);
                System.exit(-1);
            }
        }

        //ignore compiler generated constructor method
        if (isMethodPrivate && !methodName.equals("$init$")) {
            boolean isOk = true;
            try {
                //exist constructor in typedef
                checkIsMethodInTypeDef(stage2Env.getClassFile(), methodName, info, node);
                isOk = false;
            } catch (CompilerException ex) {
                //Alles inordnung nur wenn keine Exception geworfen wird ist ein Problem aufgetretten 
            }
            if (!isOk) {
                throw new CompilerException("Private method " + info + " is defined in typedef; must be declared public", node.getSymbol());
            }
        }
        try {
            stage2Env.getClassFile().addMethod(info);
        } catch (ByteCodeException ex) {
            throw new Error("Fatal Error:" + ex);
        }
        xxxEnv = null;
    }

    /**
     * check and modifie the modifer list
     *
     * @param node
     * @param outFlagList modified flag list
     * @param flagList input current flag list
     * @return Pair of Boolean Boolean, if first attribute true then is method
     * public if second attribute is fals then method is false
     * @throws CompilerException
     */
    public static Pair<Boolean, Boolean> checkAccessFlagsForTypeImpl(SyntaxNode node,
            List<MethodAccessFlag> outFlagList, Collection<ModifierNode.Modifier> flagList) throws CompilerException {
        boolean isMethodPrivate = false;
        boolean isMethodPublic = false;
        Set<MethodAccessFlag> accessFlags = new HashSet<>();

        for (ModifierNode.Modifier modis : flagList) {
            switch (modis) {
                case PUBLIC:
                    if (isMethodPrivate) {
                        throw new CompilerException("Incompatible modifier private and public", node.getSymbol());
                    }
                    isMethodPublic = true;
                    accessFlags.add(MethodAccessFlag.PUBLIC);
                    break;
                case PRIVATE:
                    if (isMethodPublic) {
                        throw new CompilerException("Incompatible modifier public and private", node.getSymbol());
                    }
                    isMethodPrivate = true;
                    accessFlags.add(MethodAccessFlag.PRIVATE);
                    break;
                case STATIC:
                    accessFlags.add(MethodAccessFlag.STATIC);
                    break;
                case SYNCHRONIZED:
                    accessFlags.add(MethodAccessFlag.SYNCHRONIZED);
                    break;
                case SYNTHETIC:
                    accessFlags.add(MethodAccessFlag.SYNTHETIC);
                    break;
                case ABSTRACT:
                case FINAL:
                case NATIVE:
                case TRANSIENT:
                case VOLATILE:
                case STRICTFP:
                case PROTECTED:
                    throw new CompilerException("unsupported access flag: " + modis, node.getSymbol());
                default:
                    throw new AssertionError(modis.name());
            }
        }
        if (!accessFlags.contains(MethodAccessFlag.PUBLIC) && !accessFlags.contains(MethodAccessFlag.PRIVATE)) {
            accessFlags.add(MethodAccessFlag.PUBLIC);
            isMethodPublic = true;
        }
        outFlagList.addAll(accessFlags);
        return new Pair<>(isMethodPublic, isMethodPrivate);
    }

    /**
     *
     * @param stage2Env
     * @param returnType the return type of the method
     * @param methodName the name of the method
     * @param methodHead parameters and throws list
     * @param accessFlags the modifier
     * @param node
     * @return the generated MethodInfo
     * @throws CompilerException
     */
    public static MethodInfo createMethodInfo(Stage2Environment stage2Env, Type returnType, String methodName, Pair<List<Pair<Type, String>>, List<String>> methodHead,
            MethodAccessFlag[] accessFlags, SyntaxNode node) throws CompilerException {
        return createMethodInfo(stage2Env, returnType, methodName, methodHead, accessFlags, node, false, null);
    }

    /**
     *
     * @param stage2Env
     * @param returnType the return type
     * @param methodName the method name
     * @param methodHead formal parameters and throws list
     * @param accessFlags modifies list
     * @param node
     * @param isGeneric is generic method
     * @param genericsParameter generic parameters
     * @return the generated MethodInfo
     * @throws CompilerException
     */
    public static MethodInfo createMethodInfo(Stage2Environment stage2Env, Type returnType, String methodName, Pair<List<Pair<Type, String>>, List<String>> methodHead,
            MethodAccessFlag[] accessFlags, SyntaxNode node, boolean isGeneric, List<Type.TypeParamType> genericsParameter) throws CompilerException {
        List<Pair<Type, String>> l = methodHead.value1;
        List<Type> parameterList = new ArrayList<>();
        //build parameter list from methodHead
        for (Pair<Type, String> pair : l) {
            if (pair.value1 instanceof Type.ArrayType) {
                MethodAccessFlag[] temp = new MethodAccessFlag[accessFlags.length + 1];
                System.arraycopy(accessFlags, 0, temp, 1, accessFlags.length);
                temp[0] = MethodAccessFlag.VARARGS;
                accessFlags = temp;
            }
            parameterList.add(pair.value1);
        }

        Type.MethodType methodType;
        List<Type> throwsList = new ArrayList<>();
        //check is generic an generate method type
        if (isGeneric) {
            methodType = new Type.MethodType(genericsParameter, parameterList, returnType, throwsList);
        } else {
            methodType = new Type.MethodType(null, parameterList, returnType, throwsList);
        }
        //build method descriptor from method type
        MethodDescriptor makeMethodDescriptor = stage2Env.getTypeUtils().makeMethodDescriptor(methodType);
        try {
            //build MethodInfo from method descriptor
            MethodInfo minfo = new MethodInfo(methodName, makeMethodDescriptor,
                    stage2Env.getClassFile().getThis(), accessFlags);
            //stage2Env.getClassFile().addMethod(minfo);
            //add signature to MethodInfo, this reprents the generic parameters
            minfo.addAttribute(new Signature(SignatureEncoder.encode(methodType)));
            //check exist throws list
            if (!methodHead.value2.isEmpty()) {
                //add throws list to MethodInfo
                Exceptions exceptions = new Exceptions();
                ClassFileRef throwable = ClassFinder.instance().find("java/lang/Throwable");
                for (String e : methodHead.value2) {
                    String qid = stage2Env.getSymbolFacade().getFullQualifiedId(e);
                    if (qid == null) {
                        qid = e;
                    }
                    exceptions.addException(new ClassInfo(qid));
                    ClassFileRef ref = ClassFinder.instance().find(qid);
                    if (ref == null) {
                        throw new CompilerException("Class not found: " + qid, node.getSymbol());
                    }
                    if (!DerivationTree.instance().isAssignable(throwable, ref)) {
                        throw new CompilerException("Expecting java.lang.Throwable", node.getSymbol());
                    }

                    throwsList.add(new Type.ClassType(null, qid, null));

                }
                //add the exception attribute to MethodInfo
                minfo.addAttribute(exceptions);
            }
            // minfo.getCode().getInstructionList().add(new Return());
            return minfo;

        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return null;
    }

    /**
     *
     * @param classFile
     * @param methodName method name
     * @param methodInfo
     * @param node
     * @throws CompilerException
     */
    public static void checkExistMethodInClassFile(ClassFile classFile, String methodName,
            MethodInfo methodInfo, SyntaxNode node) throws CompilerException {
        //get all MethodInfo from class file with the same name 
        // this are the possible candidates
        List<MethodInfo> typeImplMethods = classFile.getMethods(methodName);
        if (typeImplMethods != null) {
            for (MethodInfo mInfo : typeImplMethods) {
                MethodDescriptor mDescriptor = mInfo.getDescriptor();
                //ignore syntetic and bridged methods
                if (!mInfo.isSynthetic() && !mInfo.isBridge()) {
                    MethodDescriptor iDescriptor = methodInfo.getDescriptor();
                    if (iDescriptor.getParameters().equals(mDescriptor.getParameters())) {
                        throw new CompilerException("Method " + mInfo.toString() + " already defined", node.getSymbol());
                    }
                }

            }
        }
        //if (mInfo.getDescriptor().equals(methodInfo.getDescriptor())) {
        //    throw new CompilerException("Method " + mInfo.toString() + " already defined", node.getSymbol());
        //}
    }

    /**
     *
     * @param classFile
     * @param methodName
     * @param info
     * @param node
     * @throws CompilerException
     */
    public static void checkIsMethodInTypeDef(ClassFile classFile, String methodName, MethodInfo info, SyntaxNode node) throws CompilerException {
        try {
            try {
                //exist method in typedef
                checkMethod(ClassFinder.instance().find(classFile.getSuper().getClassName()).get(), methodName, info, node);
            } catch (CompilerException ex) {
                //check exist method in a interface of typedef
                if (interfaceCheck(ClassFinder.instance().find(classFile.getSuper().getClassName()).get(), methodName, info, node, ex)) {
                    throw ex;
                }
            }
        } catch (ByteCodeException | IOException | ResolveException e) {
            e.printStackTrace(System.err);
            System.exit(-1);
        }
    }

    /**
     *
     * @param typedef
     * @param methodName
     * @param minfo
     * @param node
     * @param ex
     * @return true exist method otherwise else
     * @throws ByteCodeException
     * @throws CompilerException
     * @throws IOException
     * @throws ResolveException
     */
    private static boolean interfaceCheck(ClassFile typedef, String methodName, MethodInfo minfo, SyntaxNode node, CompilerException ex) throws ByteCodeException, CompilerException, IOException, ResolveException {
        //iterate over all interfaces from typedef
        for (ClassInfo iface : typedef.getInterfaces()) {
            ClassFile file = ClassFinder.instance().find(iface.getClassName()).get();
            try {
                //check exist method in interface
                checkMethod(file, methodName, minfo, node);
                return false;
            } catch (CompilerException e) {
                //iterate over all interfaces of the tested interfaces
                for (ClassInfo info : file.getInterfaces()) {
                    if (!interfaceCheck(ClassFinder.instance().find(info.getClassName()).get(), methodName, minfo, node, e)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    /**
     * Check exist method in typedef
     *
     * @param classFile
     * @param methodName
     * @param info
     * @param node
     * @throws CompilerException
     */
    private static void checkMethod(ClassFile classFile, String methodName, MethodInfo info, SyntaxNode node) throws CompilerException {
        try {
            ClassFile typeDefFile = ClassFinder.instance().find(classFile.getThis().getClassName()).get();
            boolean exists = false;
            //iterate from typedef class file to java/lang/Object
            while (!typeDefFile.getThis().getClassName().equals("java/lang/Object")) {

                List<MethodInfo> methodList = new ArrayList<>();
                if (methodName.equals("<init>")) {
                    List<MethodInfo> lst = typeDefFile.getMethods("<init>");
                    if (lst != null) {
                        for (MethodInfo m : lst) {
                            if (m.isPublic()) {
                                methodList.add(m);
                            }
                        }
                    }
                } else {
                    //Find all Methods
                    Map<MethodInfo, MethodFinder.DistanceTailPair> findAll = MethodFinder.findAll(methodName, info.getDescriptor().getParameters(), typeDefFile.getThis().getClassName(),
                            typeDefFile.getThis().getClassName(), node, MethodFinder.QueryType.ALL, false);
                    if (findAll != null) {
                        for (Entry<MethodInfo, MethodFinder.DistanceTailPair> entry : findAll.entrySet()) {
                            if (entry.getValue().tailLength == 0) {
                                methodList.add(entry.getKey());
                            }
                        }
                    }
                }
                //generic testing is possible that generic method is equal to the testing method
                if (!methodList.isEmpty()) {
                    for (MethodInfo minfo : methodList) {

                        Type.MethodType methodSignatureByEnv1 = TypeUtils.getMethodSignature(minfo);
                        Type.MethodType methodSignatureByEnv2 = TypeUtils.getMethodSignature(info);


                        if (xxxEnv != null) {
                            methodSignatureByEnv1 = xxxEnv.getTypeUtils().getMethodSignatureByEnv(minfo, xxxEnv.getTypeUtils().createThis(xxxEnv.getClassFile().getThis().getClassName()), null);
                            methodSignatureByEnv2 = xxxEnv.getTypeUtils().getMethodSignatureByEnv(info, xxxEnv.getTypeUtils().createThis(xxxEnv.getClassFile().getThis().getClassName()), null);
                        }
                        String str1 = SignatureEncoder.encode(methodSignatureByEnv1);
                        String str2 = SignatureEncoder.encode(methodSignatureByEnv2);
                        if (methodSignatureByEnv1.throwsTypes != null && methodSignatureByEnv2.throwsTypes != null) {
                            List<Type> oldThrows1 = new LinkedList<>();
                            oldThrows1.addAll(methodSignatureByEnv1.throwsTypes);
                            List<Type> oldThrows2 = new LinkedList<>();
                            oldThrows2.addAll(methodSignatureByEnv2.throwsTypes);

                            methodSignatureByEnv1.throwsTypes.clear();
                            methodSignatureByEnv2.throwsTypes.clear();
                            String ostr1 = SignatureEncoder.encode(methodSignatureByEnv1);
                            String ostr2 = SignatureEncoder.encode(methodSignatureByEnv2);
                            if (!str1.equals(str2) && ostr1.equals(ostr2)) {
                                if (oldThrows1.size() == oldThrows2.size()) {
                                    Iterator<? extends Type> it1 = oldThrows1.iterator();
                                    Iterator<? extends Type> it2 = oldThrows2.iterator();
                                    Set<String> s1 = new HashSet<>();
                                    Set<String> s2 = new HashSet<>();
                                    while (it1.hasNext() && it2.hasNext()) {
                                        s1.add(it1.next().toString());
                                        s2.add(it2.next().toString());
                                    }
                                    if (!s1.equals(s2)) {
                                        throw new CompilerException("Method " + methodName + ": throws list differs from typedef", node.getSymbol());
                                    } else {
                                        //throw new CompilerException("Method " + methodName + ": throws list is in wrong order", node.getSymbol());
                                    }
                                    str1 = ostr1;
                                    str2 = ostr2;
                                } else {
                                    throw new CompilerException("Method " + methodName + ": throws list differs from typedef", node.getSymbol());
                                }
                            }
                        } else if (methodSignatureByEnv1.throwsTypes != null || methodSignatureByEnv2.throwsTypes != null) {
                            throw new CompilerException("Method " + methodName + ": throws list differs from typedef", node.getSymbol());
                        }


                        if (str1.equals(str2)) {
                            /*NOW check AccessFlags*/
                            Collection<MethodAccessFlag> tmp1 = info.getAccessFlags();
                            Collection<MethodAccessFlag> typeImplFlags = new LinkedList<>();
                            Collection<MethodAccessFlag> tmp2 = minfo.getAccessFlags();
                            Collection<MethodAccessFlag> typeDefFlags = new LinkedList<>();
                            typeDefFlags.addAll(tmp2);
                            typeImplFlags.addAll(tmp1);
                            typeImplFlags.remove(MethodAccessFlag.SYNCHRONIZED);

                            typeDefFlags.remove(MethodAccessFlag.ABSTRACT);
                            for (MethodAccessFlag flag : typeImplFlags) {
                                if (!typeDefFlags.contains(flag)) {
                                    throw new CompilerException("Method " + methodName + ": "
                                            + "flags missmatch with typedef: " + flag, node.getSymbol());
                                }
                            }

                            for (MethodAccessFlag flag : typeDefFlags) {
                                if (!typeImplFlags.contains(flag)) {
                                    throw new CompilerException("Method " + methodName + ": "
                                            + "flags missmatch with typedef: " + flag, node.getSymbol());
                                }
                            }

                            exists = true;
                            break;
                        }
                    }
                }
                if (methodName.equals("<init>") || exists) {
                    break;
                }
                typeDefFile = ClassFinder.instance().find(typeDefFile.getSuper().getClassName()).get();

            }
            if (!exists) {
                throw new CompilerException(info + " does not exist in typedef", node.getSymbol());
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            throw new Error("Fatal error:" + ex.getMessage());
        }
    }

    /**
     * Create a generic method head
     *
     * @param node
     * @param stage2Env
     * @param returnType the return type of the method
     * @param methodName the name of the method
     * @param methodHead formal parametrs and throw list
     * @param flagList method modifier
     * @param typeParams generic parameters
     * @throws CompilerException
     */
    public static void buildGenericMethodHead(SyntaxNode node,
            Stage2Environment stage2Env, Type returnType, String methodName,
            Pair<List<Pair<Type, String>>, List<String>> methodHead,
            List<ModifierNode.Modifier> flagList, List<Type.TypeParamType> typeParams) throws CompilerException {
        //xxxEnv = stage2Env;
        MethodAccessFlag[] accessFlags;
        List<MethodAccessFlag> outList = new LinkedList<>();
        //check and modifie the access flags
        Pair<Boolean, Boolean> methodAccess = MethodHeadBuilder.checkAccessFlagsForTypeImpl(node, outList, flagList);
        boolean isMethodPublic = methodAccess.value1;
        boolean isMethodPrivate = methodAccess.value2;
        accessFlags = outList.toArray(new MethodAccessFlag[0]);

        //build method info
        MethodInfo info = createMethodInfo(stage2Env, returnType, methodName, methodHead, accessFlags, node, true, typeParams);
        //store method info
        stage2Env.pushMethodInfo(info);
        //check exist method in class file
        checkExistMethodInClassFile(stage2Env.getClassFile(), methodName,
                info, node);
        if (isMethodPublic) {
            //if method is public 
            //then the method must be exists in typedef or base type or interface
            checkIsMethodInTypeDef(stage2Env.getClassFile(), methodName, info, node);
        }
        if (isMethodPrivate) {
            boolean isOk = true;
            try {
                //if method is private 
                //then the method doesn't  exists in typedef or base type or interface 
                checkIsMethodInTypeDef(stage2Env.getClassFile(), methodName, info, node);
                isOk = false;
            } catch (CompilerException ex) {
                //Alles inordnung nur wenn keine Exception geworfen wird ist ein Problem aufgetretten 
            }
            if (!isOk) {
                throw new CompilerException("Private method " + info + " is defined in typedef; must be declared public", node.getSymbol());
            }
        }
        try {
            stage2Env.getClassFile().addMethod(info);
        } catch (ByteCodeException ex) {
            throw new Error("Fatal Error:" + ex);
        }
    }

    /**
     *
     * @param formalParametersVisitor
     * @param throwsVisitor
     * @return the generates Pair of parameter and throw list
     * @throws CompilerException
     */
    public static Pair<List<Pair<Type, String>>, List<String>> makeMethodHead(Stage2FormalParametersVisitor formalParametersVisitor,
            QualifiedIdentifierListVisitor throwsVisitor) throws CompilerException {
        List<String> throwsList;
        List<Pair<Type, String>> formalParameters = formalParametersVisitor.getFieldContainer();
        if (throwsVisitor == null) {
            throwsList = Collections.EMPTY_LIST;
        } else {
            throwsList = throwsVisitor.getQualifiedIdentifiers();
        }
        if (formalParameters == null) {
            formalParameters = Collections.EMPTY_LIST;
        }

        return new Pair<>(formalParameters, throwsList);

    }

    /**
     *
     * @param stage2Env
     * @param methodAccessFlag
     * @param returnType
     * @param methodName
     * @param methodHead
     * @param node
     * @param typeParams
     * @param isGeneric
     * @throws CompilerException
     */
    public static void buildInterfaceMethodHead(Stage2Environment stage2Env, MethodAccessFlag[] methodAccessFlag, Type returnType, String methodName, Pair<List<Pair<Type, String>>, List<String>> methodHead, SyntaxNode node, List<Type.TypeParamType> typeParams, boolean isGeneric) throws CompilerException {
        MethodInfo createMethodInfo = MethodHeadBuilder.createMethodInfo(stage2Env, returnType, methodName, methodHead, methodAccessFlag, node, isGeneric, typeParams);
        MethodHeadBuilder.checkExistMethodInClassFile(stage2Env.getClassFile(), methodName, createMethodInfo, node);
        try {
            stage2Env.getClassFile().addMethod(createMethodInfo);
        } catch (ByteCodeException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }

    }

    public static MethodAccessFlag[] getMethodAccessFlags(String methodName, Collection<ModifierNode> modis) throws CompilerException {
        MethodAccessFlag[] flags = new MethodAccessFlag[modis.size()];
        boolean[] uniqueAccess = {false, false, false,
            false, false, false,
            false, false, false};
        int i = 0;
        for (ModifierNode mod : modis) {
            switch (mod.getModifier()) {
                case ABSTRACT:
                    if (!uniqueAccess[0]) {
                        flags[i] = MethodAccessFlag.ABSTRACT;
                        uniqueAccess[0] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case FINAL:
                    if (!uniqueAccess[1]) {
                        flags[i] = MethodAccessFlag.FINAL;
                        uniqueAccess[1] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case STATIC:
                    if (!uniqueAccess[2]) {
                        flags[i] = MethodAccessFlag.STATIC;
                        uniqueAccess[2] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case SYNCHRONIZED:
                    if (!uniqueAccess[3]) {
                        flags[i] = MethodAccessFlag.SYNCHRONIZED;
                        uniqueAccess[3] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case PUBLIC:
                    if (!uniqueAccess[4]) {
                        flags[i] = MethodAccessFlag.PUBLIC;
                        uniqueAccess[4] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case PROTECTED:
                    if (!uniqueAccess[5]) {
                        flags[i] = MethodAccessFlag.PROTECTED;
                        uniqueAccess[5] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case PRIVATE:
                    if (!uniqueAccess[6]) {
                        flags[i] = MethodAccessFlag.PRIVATE;
                        uniqueAccess[6] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case NATIVE:
                    if (!uniqueAccess[7]) {
                        flags[i] = MethodAccessFlag.NATIVE;
                        uniqueAccess[7] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                case STRICTFP:
                    if (!uniqueAccess[8]) {
                        flags[i] = MethodAccessFlag.STRICT;
                        uniqueAccess[8] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + methodName, mod.getSymbol());
                    }
                    break;
                default:
                    throw new AssertionError("Du kannst mich mal");
            }
            i++;
        }

        return flags;

    }

    public static FieldAccessFlag[] getFieldAccessFlags(String fieldName, Collection<ModifierNode> modis) throws CompilerException {
        FieldAccessFlag[] flags = new FieldAccessFlag[modis.size()];
        boolean[] uniqueAccess = {false, false, false,
            false, false, false,
            false, false, false};
        int i = 0;
        for (ModifierNode mod : modis) {
            switch (mod.getModifier()) {

                case FINAL:
                    if (!uniqueAccess[1]) {
                        flags[i] = FieldAccessFlag.FINAL;
                        uniqueAccess[1] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + fieldName, mod.getSymbol());
                    }
                    break;
                case STATIC:
                    if (!uniqueAccess[2]) {
                        flags[i] = FieldAccessFlag.STATIC;
                        uniqueAccess[2] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + fieldName, mod.getSymbol());
                    }
                    break;

                case PUBLIC:
                    if (!uniqueAccess[4]) {
                        flags[i] = FieldAccessFlag.PUBLIC;
                        uniqueAccess[4] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + fieldName, mod.getSymbol());
                    }
                    break;
                case PROTECTED:
                    if (!uniqueAccess[5]) {
                        flags[i] = FieldAccessFlag.PROTECTED;
                        uniqueAccess[5] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + fieldName, mod.getSymbol());
                    }
                    break;
                case PRIVATE:
                    if (!uniqueAccess[6]) {
                        flags[i] = FieldAccessFlag.PRIVATE;
                        uniqueAccess[6] = true;
                    } else {
                        throw new CompilerException("Dublicate modifier at " + fieldName, mod.getSymbol());
                    }
                    break;


                default:
                    throw new CompilerException("Unsupported modifier " + mod + " for field " + fieldName, mod.getSymbol());
            }
            i++;
        }
        return flags;

    }

    public static void buildMethodHeadClass(SyntaxNode node, Stage2Environment stage2Env, Type type,
            String methodName, Pair<List<Pair<Type, String>>, List<String>> methodHead, MethodAccessFlag[] accessFlags) throws CompilerException {
        MethodInfo info = createMethodInfo(stage2Env, type, methodName, methodHead, accessFlags, node);
        stage2Env.pushMethodInfo(info);
        checkExistMethodInClassFile(stage2Env.getClassFile(), methodName, info, node);

        try {
            stage2Env.getClassFile().addMethod(info);
        } catch (ByteCodeException ex) {
            throw new Error("Fatal Error:" + ex);
        }
    }

    public static void buildGenericMethodHeadClass(SyntaxNode node, Stage2Environment stage2Env, Type returnType, String methodName, Pair<List<Pair<Type, String>>, List<String>> methodHead, MethodAccessFlag[] methodAccessFlags, List<Type.TypeParamType> typeParams) throws CompilerException {
        MethodInfo info = createMethodInfo(stage2Env, returnType, methodName, methodHead, methodAccessFlags, node, true, typeParams);
        stage2Env.pushMethodInfo(info);
        checkExistMethodInClassFile(stage2Env.getClassFile(), methodName, info, node);

        try {
            stage2Env.getClassFile().addMethod(info);
        } catch (ByteCodeException ex) {
            throw new Error("Fatal Error:" + ex);
        }
    }
}
