/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package stage3;

import abstractsyntaxtree.ExpressionNode;
import abstractsyntaxtree.InfixOpNode;
import codegeneration.Visitor;
import codegeneration.abstractvisitors.AbstractAssertStatementVisitor;
import codegeneration.environment.ExpressionEnvironment;
import codegeneration.environment.Stage3Environment;
import codegeneration.instruction.InstructionCreator;
import exceptions.CompilerException;
import java.util.Deque;
import java.util.LinkedList;
import stage3.states.BlockState;
import teajay.bytecode.attributes.code.AThrow;
import teajay.bytecode.attributes.code.Dup;
import teajay.bytecode.attributes.code.Dup2;
import teajay.bytecode.attributes.code.InvokeSpecial;
import teajay.bytecode.attributes.code.Label;
import teajay.bytecode.attributes.code.NOP;
import teajay.bytecode.attributes.code.New;
import teajay.bytecode.attributes.code.Pop;
import teajay.bytecode.attributes.code.Swap;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.cpool.MethodRefInfo;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.generics.parser.Type;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
public class Stage3AssertStatementVisitor extends AbstractAssertStatementVisitor {

    private final Stage3Environment stage3Env;
    private Deque<InfixOpNode.InfixOps> stack;
    private Label label;
    private final boolean hasColon;
    private boolean firstRun;

    public Stage3AssertStatementVisitor(Stage3Environment stage3Env, boolean hasColon) {
        this.stage3Env = stage3Env;
        this.hasColon = hasColon;
        this.firstRun = hasColon;
    }

    @Override
    public void begin(ExpressionNode node) throws CompilerException {
        stack = new LinkedList<>();
        stage3Env.createNewExpressionEnvironment();
    }

    @Override
    public Visitor visit(ExpressionNode node) throws CompilerException {
        return new Stage3ExpressionVisitor(stage3Env, this.stack, true);
    }

    @Override
    public void end(ExpressionNode node) throws CompilerException {
        ExpressionEnvironment argEnv = stage3Env.getExpressionEnvironment();
        InstructionCreator.PutOperatorInstructions(argEnv, stack, node);
        if (argEnv.isShortCircuit()) {
            Label gotoLabel = new Label("gotoLabel");

            argEnv.writeShortCircuit(gotoLabel);
            BlockState hasState = stage3Env.hasState(BlockState.class);
            hasState.addLabel(gotoLabel);

            argEnv.insertCodeWithVerificationStack(new NOP(), argEnv.getOperandStack());
        }
        if (!firstRun) {
            try {
                if (hasColon) {
                    argEnv.insertCode(new Swap());
                    Type t1 = argEnv.popType();
                    Type t2 = argEnv.popType();
                    argEnv.setResultType(t1);
                    if ((t1 instanceof Type.SimpleType) && ((Type.SimpleType) t1).isPrimitiveType()) {
                        throw new CompilerException("incompatible type for assert: " + t1);
                    }
                    argEnv.setResultType(t2);
                }

                Type typeOfResult = argEnv.typeOfResult();
                if (typeOfResult instanceof Type.ClassType) {
                    if ((((Type.ClassType) typeOfResult).name).equals("java/lang/Boolean")) {
                        this.label = argEnv.writeNeqIf();
                        argEnv.popType();
                        argEnv.insertCode(new New(ClassInfo.fromClass(AssertionError.class)));
                        if (hasColon) {
                            argEnv.insertCode(new Dup2());
                            argEnv.insertCode(new Swap());
                        } else {
                            argEnv.insertCode(new Dup());
                        }

                        if (hasColon) {
                            argEnv.insertCode(new InvokeSpecial(MethodRefInfo.make(
                                    ClassInfo.fromClass(AssertionError.class),
                                    "<init>", FieldDescriptor.getVoid(), FieldDescriptor.fromClass(Object.class))));
                        } else {
                            argEnv.insertCode(new InvokeSpecial(MethodRefInfo.make(
                                    ClassInfo.fromClass(AssertionError.class),
                                    "<init>", FieldDescriptor.getVoid())));
                        }
                        argEnv.insertCode(new AThrow());
                        BlockState state = stage3Env.hasState(BlockState.class);
                        state.addLabel(label);
                        if (hasColon) {
                            argEnv.insertCodeWithVerificationStack(new Pop(), argEnv.getOperandStack());
                            argEnv.popType();
                        } else {
                            argEnv.insertCode(new NOP());
                        }

                    } else {
                        throw new CompilerException("assert statement expects Boolean expression", node.getSymbol());
                    }
                } else {
                    throw new CompilerException("assert statement expects Boolean expression", node.getSymbol());
                }
            } catch (ByteCodeException ex) {
                ex.printStackTrace(System.err);
                System.exit(-1);
            }
        } else {
            firstRun = false;
        }
    }

    public Label getLabel() {
        return null;
    }
}
