/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package symbol;

import teajay.runtime.exceptions.ResolveException;
import java.io.IOException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.logging.Logger;
import teajay.bytecode.ClassFile;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;

/**
 *
 * @author Michael (Burzan | Kußmann)
 */
public class DerivationTree {

    private static DerivationTree instance = null;
    private final HashMap<ClassFileRef, DerivationNode> map = new HashMap<>(1024);
    private DerivationNode root; //Object;

    public static DerivationTree instance() throws ByteCodeException, ResolveException, IOException {
        if (instance == null) {
            instance = new DerivationTree();
        }
        return instance;
    }

    public int getDepth(ClassFileRef ref) throws ByteCodeException, IOException, ResolveException {
        if (ref.referencesArray()) {
            FieldDescriptor a = FieldDescriptor.fromString(ref.getQID());
            ClassInfo leafType = a.getLeafElementType().getClassInfo();
            /*if (leafType.isPrimitive()) {
                return 0;
            }*/
            ClassFileRef aLeafRef = ClassFinder.instance().find(leafType.getClassName());
            if (aLeafRef == null) {
                throw new ResolveException("Class not found: " + ref.getQID());
            }
            return getDepth(aLeafRef) + a.getDimension();
        } else if (ref.referencesNull()) {
            return 0;
        } else if (ref.referencesPrimitive()) {
            return 0;
        }
        DerivationNode result = map.get(ref);
        if (result == null) {
            result = insertClassDerivation(ref);
        }
        return result.getDepth();
    }

    //root node is the class file of Object
    private DerivationTree() throws ByteCodeException, ResolveException, IOException {
        ClassFinder finder = ClassFinder.instance();
        ClassFileRef object = finder.find("java/lang/Object");
        if (object != null) {
            this.root = new DerivationNode(null, object, 0);
            this.map.put(object, root);
        } else {
            throw new Error("java/lang/Object not found. Is the ClassFinder initialized properly?");
        }
    }

    private DerivationNode insertClassDerivation(ClassFileRef ref) throws ByteCodeException, IOException, ResolveException {
        ClassFinder finder = ClassFinder.instance();
        DerivationNode refnode = map.get(ref);
        if (ref != this.root.getClassFileRef() && refnode == null) {
            LinkedList<ClassFileRef> superClasses = new LinkedList<>();
            ClassFileRef s;
            //Handle interfaces properly: interfaces !implement! the interface they are extending.
            if (ref.get().isInterface()) {
                Collection<ClassInfo> ifaces = ref.get().getInterfaces();
                if (ifaces.size() == 1) {
                    s = ClassFinder.instance().find(ifaces.iterator().next().getClassName());
                } else {
                    s = finder.find(ref.get().getSuper().getClassName());
                }
            } else {
                s = finder.find(ref.get().getSuper().getClassName());
            }

            while (s != this.root.getClassFileRef()) {
                superClasses.add(s);
                if (s.get().isInterface()) {
                    Collection<ClassInfo> ifaces = s.get().getInterfaces();
                    if (ifaces.size() == 1) {
                        s = ClassFinder.instance().find(ifaces.iterator().next().getClassName());
                    } else {
                        s = finder.find(s.get().getSuper().getClassName());
                    }
                } else {
                    s = finder.find(s.get().getSuper().getClassName());
                }
            }
            Iterator<ClassFileRef> iteraror = superClasses.descendingIterator();
            DerivationNode node;
            if (iteraror.hasNext()) {
                ClassFileRef r = iteraror.next();
                if (!this.map.containsKey(r)) {
                    node = new DerivationNode(root, r, 1);
                    root.getChildList().add(node);
                    map.put(r, node);
                } else {
                    node = map.get(r);
                }

                while (iteraror.hasNext()) {
                    r = iteraror.next();
                    if (!this.map.containsKey(r)) {
                        DerivationNode nextNode = new DerivationNode(node, r, node.getDepth() + 1);
                        node.getChildList().add(nextNode);
                        node = nextNode;
                        map.put(r, node);
                    } else {
                        node = map.get(r);
                    }
                }

            } else {
                node = root;
            }
            refnode = new DerivationNode(node, ref, node.getDepth() + 1);
            node.getChildList().add(refnode);
            map.put(ref, refnode);
        }
        return refnode;
    }

    @Deprecated
    public LinkedList<ClassFileRef> getSuperClasses(ClassFileRef ref) throws ByteCodeException, IOException, ResolveException {
        if (ref.referencesNull() || ref.referencesPrimitive()) {
            return new LinkedList<>();
        } else if (ref.referencesArray()) {
            FieldDescriptor a = FieldDescriptor.fromString(ref.getQID());
            ClassInfo leafType = a.getLeafElementType().getClassInfo();
            ClassFileRef find = ClassFinder.instance().find(leafType.getClassName());
            LinkedList<ClassFileRef> result = new LinkedList<>();
            for (ClassFileRef cr : getSuperClasses(find)) {
                FieldDescriptor descr = FieldDescriptor.makeArray(FieldDescriptor.fromString(cr.getQID()), a.getDimension());
                result.add(ClassFinder.instance().find(descr.getClassInfo().getClassName()));
            }
            return result;
        }
        if (map.containsKey(ref)) {
            LinkedList<ClassFileRef> list = new LinkedList<>();
            DerivationNode begin = map.get(ref);
            if (begin == root) {
                list.add(root.getClassFileRef());
                return list;
            }
            begin = begin.getParent();
            while (begin != root) {
                list.add(begin.getClassFileRef());
                begin = begin.getParent();
            }
            list.add(root.getClassFileRef());
            return list;
        } else {
            insertClassDerivation(ref);
            return getSuperClasses(ref);
        }
    }

    public boolean isAssignable(ClassFileRef base, ClassFileRef derivate) throws ByteCodeException, IOException, ResolveException {
        if (base.referencesArray() && derivate.referencesArray()) {
            FieldDescriptor a = FieldDescriptor.fromString(base.getQID());
            FieldDescriptor b = FieldDescriptor.fromString(derivate.getQID());
            ClassFileRef aLeafRef = ClassFinder.instance().find(a.getLeafElementType().getClassInfo().getClassName());
            ClassFileRef bLeafRef = ClassFinder.instance().find(b.getLeafElementType().getClassInfo().getClassName());
            if (aLeafRef == null || bLeafRef == null) {
                throw new ResolveException("Class not found: " + (aLeafRef == null ? base.getQID() : derivate.getQID()));
            }
            if (a.getDimension() == b.getDimension()) {
                return isAssignable(aLeafRef, bLeafRef);
            } else if (aLeafRef.getQID().equals("java/lang/Object") && bLeafRef.getQID().equals("java/lang/Object")) {
                return a.getDimension() < b.getDimension();
            }
            return false;
        } else if (derivate.referencesArray()) {
            return base == root.getClassFileRef(); //Arrays extend Object
        } else if (base.referencesArray()) {
            return false;
        } else if (base.referencesPrimitive() && derivate.referencesPrimitive()) {
            return base == derivate;
        } else if (base.referencesPrimitive() || derivate.referencesPrimitive()) {
            return false;
        }

        if (isDerivated(base, derivate)) {
            return true;
        }
        ClassFile left = base.get();
        if (!left.isInterface()) {
            return false;
        }
        ClassFile right = derivate.get();
        Collection<ClassInfo> rightInterfaces = right.getInterfaces();
        ClassFileRef leftRef = base;
        for (ClassInfo rightInterface : rightInterfaces) {
            ClassFileRef rightRef = ClassFinder.instance().find(rightInterface.getClassName());
            if (rightRef == null) {
                throw new ResolveException("Class not found: " + rightInterface.getClassName());
            }
            if (isDerivated(leftRef, rightRef)) {
                return true;
            }
        }

        if (right.getSuper() != null) {
            ClassFileRef superClass = ClassFinder.instance().find(right.getSuper().getClassName());
            if (superClass == null) {
                throw new ResolveException("Class not found: " + right.getSuper().getClassName());
            }
            return isAssignable(base, superClass);
        }
        return false;
    }

    public boolean isDerivated(ClassFileRef base, ClassFileRef derivate) throws ByteCodeException, IOException, ResolveException {

        if (base.referencesPrimitive() || derivate.referencesPrimitive()) {
            return false;
        } else if (derivate.referencesNull()) {
            return true;
        }

        if (base.referencesArray() && derivate.referencesArray()) {
            FieldDescriptor a = FieldDescriptor.fromString(base.getQID());
            FieldDescriptor b = FieldDescriptor.fromString(derivate.getQID());
            if (a.getDimension() == b.getDimension()) {
                ClassFileRef refa = ClassFinder.instance().find(a.getLeafElementType().getClassInfo().getClassName());
                ClassFileRef refb = ClassFinder.instance().find(b.getLeafElementType().getClassInfo().getClassName());
                if (refa == null || refb == null) {
                    throw new ResolveException("Class not found: " + (refa == null ? base.getQID() : derivate.getQID()));
                }
                return isDerivated(refa, refb);
            } else {
                return false;
            }
        } else if (derivate.referencesArray()) {
            return (base == root.getClassFileRef()); //Object 
        } else if (base.referencesArray()) {
            return false;
        }
        //Fix: Dieselbe Klasse wird nicht als verwandt eingestuft
        if (base == derivate) {
            return true;
        }

        if (!this.map.containsKey(base)) {
            insertClassDerivation(base);
        }
        if (!this.map.containsKey(derivate)) {
            insertClassDerivation(derivate);
        }


        DerivationNode baseNode = map.get(base);
        DerivationNode derivateNode = map.get(derivate);
        if (baseNode.getDepth() == derivateNode.getDepth()) {
            return false;
        }

        //if baseNode == java/lang/Object
        if (baseNode == root) {
            return true;
        }
        DerivationNode iter = derivateNode.getParent();
        while (iter != null) {
            if (iter == baseNode) {
                return true;
            }
            iter = iter.getParent();
        }
        return false;
    }
    
    public boolean isAssignableAndDifferent(ClassFileRef ref1, ClassFileRef ref2) throws ByteCodeException, IOException, ResolveException {
        if(ref1.get().getThis().getClassName().equals(ref2.get().getThis().getClassName())){
            return false;
        }
        return isAssignable(ref1, ref2);
    }
    

    public void print() {
        System.out.println("\n\n");
        root.print();
    }
}

class DerivationNode {

    private ClassFileRef classFileRef;
    private DerivationNode parent;
    private LinkedList<DerivationNode> childList = new LinkedList<>();
    private int depth;

    DerivationNode(DerivationNode parent, ClassFileRef classFileRef, int depth) {
        this.parent = parent;
        this.classFileRef = classFileRef;
        this.depth = depth;
    }

    public ClassFileRef getClassFileRef() {
        return classFileRef;
    }

    public DerivationNode getParent() {
        return parent;
    }

    public LinkedList<DerivationNode> getChildList() {
        return childList;
    }

    public int getDepth() {
        return depth;
    }

    public void print() {
        print("\t");
    }

    private void print(String tab) {
        try {
            System.out.println("Name:" + this.classFileRef.get().getThis().getClassName());
            System.out.println("Depth:" + this.getDepth() + "\n===============================");
            for (DerivationNode n : childList) {
                n.print();
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            Logger.getLogger(DerivationNode.class.getName()).log(Level.SEVERE, null, ex);
        }
    }
}
