/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package symbol;

import java.util.Arrays;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import teajay.bytecode.attributes.code.CodeAttribute;
import teajay.bytecode.attributes.code.JVMInstruction;
import teajay.bytecode.attributes.code.StackMapTable;
import teajay.bytecode.attributes.code.VerificationType;
import teajay.bytecode.methods.MethodInfo;
import teajay.generics.check.TypeUtils;
import teajay.generics.parser.Type;

/**
 *
 * @author Michael (Burzan | Kußmann )
 */
public final class LocalTable implements Iterable<Frame> {

    private final Deque<Frame> stack = new LinkedList<>();
    private final StackMapGenerator gen;
    private final TypeUtils typeUtils;
    private final MethodInfo methodInfo;

    public LocalTable(MethodInfo info) {
        this(info, new TypeUtils());
    }

    public LocalTable(MethodInfo methodInfo, TypeUtils typeUtils) {
        CodeAttribute code = methodInfo.getCode();
        if (code != null) {
            StackMapTable table = code.getStackMapTable();
            if (table != null) {
                this.gen = new StackMapGenerator(this, table);
            } else {
                this.gen = null;
            }
        } else {
            this.gen = null;
        }
        this.typeUtils = typeUtils;
        this.methodInfo = methodInfo;
    }

    public void addFrame() {
        if (stack.isEmpty()) {
            stack.push(new Frame(0, typeUtils));
        } else {
            stack.push(new Frame(stack.peek().getMaxIndex(), typeUtils));
        }
    }

    public void addEntry(String name, Type type) {
        if (name == null || type == null) {
            throw new NullPointerException(name == null ? "name" : "type");
        }
        if (stack.isEmpty()) {
            throw new IllegalStateException("the local table doesn't have any frames");
        }
        stack.peek().addEntry(name, type);
    }

    public void removeFrame() {
        if (stack.isEmpty()) {
            throw new IllegalStateException("the local table doesn't have any frames");
        }
        stack.pop();
    }

    public LocalEntry getEntry(String name) {
        Iterator<Frame> iter = stack.iterator();
        while (iter.hasNext()) {
            Frame f = iter.next();
            LocalEntry entry = f.get(name);
            if (entry != null) {
                return entry;
            }
        }
        return null;
    }

    @Override
    public Iterator<Frame> iterator() {
        return stack.descendingIterator();
    }

    public void addStackMapFrame(JVMInstruction start, VerificationType... stack) {
        addStackMapFrame(start, Arrays.asList(stack));
    }

    public void addStackMapFrame(JVMInstruction start, List<VerificationType> stack) {
        if (gen != null) {
            gen.addStackMapFrame(start, stack);
        } else {
            throw new IllegalStateException("no stack map table available");
        }
    }

    @Override
    public String toString() {
        StringBuilder buf = new StringBuilder();
        StringBuilder tabBuf = new StringBuilder();
        buf.append("-- LocalTable ( ").append(methodInfo).append(" )\n");
        for (Frame frame : this) {
            buf.append(frame.toString(tabBuf.toString())).append('\n');
            tabBuf.append('\t');
            buf.append(tabBuf);
        }
        return buf.toString();
    }

    public int getNextId() {
        if (stack.isEmpty()) {
            throw new IllegalStateException("the local table doesn't have any frames");
        }
        return stack.peek().getMaxIndex();
    }
}
/**
 * **
 *
 * //frame: = this, 0 int func(int c){ int a ; }
 *
 */
