/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.generics.check;

import java.util.ArrayList;
import java.util.List;
import teajay.generics.parser.Type;
import static teajay.generics.parser.Type.WildcardType.Kind.EXTENDS;
import static teajay.generics.parser.Type.WildcardType.Kind.SUPER;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
class EraseVisitor implements Type.Visitor<Type, Void> {

    private static final Type object = new Type.ClassType(null, "java/lang/Object", null);
    private final Context context;

    public EraseVisitor(Context context) {
        this.context = context;
    }

    @Override
    public Type visitSimpleType(Type.SimpleType type, Void p) {
        if (type.isPrimitiveType()) {
            return type;
        } else {
            Type t = context.getErasedType(type.name);
            return t != null ? t : object;
        }
    }

    @Override
    public Type visitArrayType(Type.ArrayType type, Void p) {
        return new Type.ArrayType(type.elemType.accept(this, null));
    }

    @Override
    public Type visitMethodType(Type.MethodType type, Void p) {
        if (type.typeParamTypes != null) {
            context.addFrame();
            for (Type t : type.typeParamTypes) {
                t.accept(this, null);
            }
        }
        List<Type> paramTypes = new ArrayList<>();
        if (type.paramTypes != null) {
            for (Type t : type.paramTypes) {
                paramTypes.add(t.accept(this, null));
            }
        }
        Type returnType = type.returnType.accept(this, null);
        List<Type> throwsTypes = new ArrayList<>();
        if (type.throwsTypes != null) {
            for (Type t : type.throwsTypes) {
                throwsTypes.add(t.accept(this, null));
            }
        }
        return new Type.MethodType(null, paramTypes, returnType, throwsTypes);
    }

    @Override
    public Type visitClassSigType(Type.ClassSigType type, Void p) {
        if (type.typeParamTypes != null) {
            context.addFrame();
            for (Type t : type.typeParamTypes) {
                t.accept(this, null);
            }
        }
        Type superclassType = type.superclassType.accept(this, null);
        List<Type> superInterfaces = new ArrayList<>();
        if (type.superinterfaceTypes != null) {
            for (Type t : type.superinterfaceTypes) {
                superInterfaces.add(t.accept(this, null));
            }
        }
        return new Type.ClassSigType(null, superclassType, superInterfaces);
    }

    @Override
    public Type visitClassType(Type.ClassType type, Void p) {
        /*if (type.name.equals("teajay/lang/Array")) {
         return utils.toJavaArrayType(type);
         }*/
        Type.ClassType outerType = null;
        if (type.outerType != null) {
            outerType = (Type.ClassType) type.outerType.accept(this, null);
        }
        return new Type.ClassType(outerType, type.name, null);
    }

    @Override
    public Type visitTypeParamType(Type.TypeParamType type, Void p) {
        if (type.classBound != null) {
            Type classBound = type.classBound.accept(this, null);
            if (!(classBound instanceof Type.ClassType) || !((Type.ClassType) classBound).name.equals("java/lang/Object")) {
                context.put(type.name, classBound, type);
                return classBound;
            }
        }
        if (type.interfaceBounds != null && type.interfaceBounds.size() > 0) {
            Type classBound = type.interfaceBounds.get(0).accept(this, null);
            context.put(type.name, classBound, type);
            return classBound;
        }
        context.put(type.name, object, type);
        return object;
    }

    @Override
    public Type visitWildcardType(Type.WildcardType type, Void p) {
        switch (type.kind) {
            case EXTENDS:
                return type.boundType.accept(this, null);
            case SUPER:
            default:
                return object;
        }
    }
}
