/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.methods;

import abstractsyntaxtree.SyntaxNode;
import datastructs.Pair;
import exceptions.CompilerException;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import symbol.DerivationTree;
import teajay.bytecode.ClassFile;
import teajay.bytecode.cpool.ClassInfo;
import teajay.bytecode.cpool.FieldDescriptor;
import teajay.bytecode.exceptions.ByteCodeException;
import teajay.bytecode.methods.MethodInfo;
import teajay.runtime.exceptions.ResolveException;
import teajay.util.QIDTools;
import teajay.util.io.ClassFileRef;
import teajay.util.io.ClassFinder;
import teajay.util.io.FileIterator;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Pure method resolving, parameter types must be erased.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public class MethodFinder {

    public static class DistanceTailPair {

        public final int distance;
        public final int tailLength;

        public DistanceTailPair(int distance, int tail) {
            this.distance = distance;
            this.tailLength = tail;
        }

        @Override
        public String toString() {
            return "distance=" + distance + ", tail=" + tailLength;
        }
    }

    public enum QueryType {

        STATIC_ONLY,
        NON_VIRTUAL,
        NON_VIRTUAL_NON_STATIC,
        NON_STATIC,
        NON_STATIC_EXACT,
        ALL;
    }

    public static Map<MethodInfo, DistanceTailPair> findAll(String name, List<FieldDescriptor> parameterTypes, String location, String caller,
            SyntaxNode node, QueryType queryType, boolean ignoreSynthetic) throws CompilerException {
        Map<MethodInfo, DistanceTailPair> candidates = new HashMap<>();
        try {
            findMethod(name, parameterTypes, location, caller, node, queryType, candidates, new HashSet<ClassInfo>(), ignoreSynthetic);
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return candidates;
    }

    public static Map<MethodInfo, DistanceTailPair> findAll(String name, List<FieldDescriptor> parameterTypes, String location, String caller,
            SyntaxNode node, QueryType queryType) throws CompilerException {
        return findAll(name, parameterTypes, location, caller, node, queryType, true);
    }

    public static Pair<MethodInfo, DistanceTailPair> findMethod(String name, String location, String caller,
            SyntaxNode node, QueryType queryType, FieldDescriptor... paramTypes) throws CompilerException {
        return findMethod(name, Arrays.asList(paramTypes), location, caller, node, queryType);
    }

    public static Pair<MethodInfo, DistanceTailPair> findMethod(String name, List<FieldDescriptor> parameterTypes, String location, String caller,
            SyntaxNode node, QueryType queryType) throws CompilerException {
        return findMethod(name, parameterTypes, location, caller, node, queryType, true);
    }
    /**
     * Resolves a method.
     * @param name name of the method
     * @param parameterTypes argument types of the method
     * @param location starting point for the search (the class that defines the method)
     * @param caller the class that tries to call the method
     * @param node a syntax node for exception creation
     * @param queryType search mode
     * @param ignoreSynthetic if false: Ignores synthetic methods.
     * @return the found method, its score and tail-length
     * @throws CompilerException if the method reference is ambiguous
     */
    public static Pair<MethodInfo, DistanceTailPair> findMethod(String name, List<FieldDescriptor> parameterTypes, String location, String caller,
            SyntaxNode node, QueryType queryType, boolean ignoreSynthetic) throws CompilerException {
        try {
            Map<MethodInfo, DistanceTailPair> candidates = new HashMap<>();
            findMethod(name, parameterTypes, location, caller, node, queryType, candidates, new HashSet<ClassInfo>(), ignoreSynthetic);
            if (!candidates.isEmpty()) {
                int minDist = Integer.MAX_VALUE;
                Entry<MethodInfo, DistanceTailPair> result = null;
                boolean ambigiuous = true;
                for (Entry<MethodInfo, DistanceTailPair> m : candidates.entrySet()) {
                    int distance = m.getValue().distance;
                    if (distance < minDist) {
                        minDist = distance;
                        result = m;
                        ambigiuous = false;
                    } else if (distance == minDist) {
                        assert result != null; //May only occur for distance == Integer.MAX_VALUE
                        ClassFileRef refa = ClassFinder.instance().find(result.getKey().getDescriptor().getReturnType().getClassInfo().getClassName());
                        ClassFileRef refb = ClassFinder.instance().find(m.getKey().getDescriptor().getReturnType().getClassInfo().getClassName());
                        ClassFileRef drefa = ClassFinder.instance().find(result.getKey().getDefiningClass().getClassName());
                        ClassFileRef drefb = ClassFinder.instance().find(m.getKey().getDefiningClass().getClassName());
                        checkRefs(refa, refb, drefa, drefb);
                        if (result.getKey().getDescriptor().getParameters().equals(m.getKey().getDescriptor().getParameters())) {
                            int aDepth = DerivationTree.instance().getDepth(drefa);
                            int bDepth = DerivationTree.instance().getDepth(drefb);
                            if (aDepth < bDepth) {
                                ClassFileRef tmp = refa;
                                refa = refb;
                                refb = tmp;
                            }
                            if (DerivationTree.instance().isAssignable(refb, refa)) {
                                //OK
                                if (aDepth < bDepth) {
                                    result = m;
                                }
                            } else {
                                ambigiuous = true;
                            }
                            //Prefer non vararg methods over vararg methods (java's behaviour)
                        } else if (result.getKey().isVarArgs() && !m.getKey().isVarArgs()) {
                            result = m;
                            ambigiuous = false;
                        } else if (result.getKey().isVarArgs() || !m.getKey().isVarArgs()) {
                            ambigiuous = true;
                        }
                    }
                }
                if (ambigiuous) {
                    if (node != null) {
                        throw new CompilerException("ambiguous method reference for " + name + " " + parameterTypes, node.getSymbol());
                    } else {
                        throw new CompilerException("ambiguous method reference for " + name + " " + parameterTypes);
                    }
                }
                assert result != null;
                return new Pair<>(result.getKey(), result.getValue());
            }
        } catch (ByteCodeException | IOException | ResolveException ex) {
            ex.printStackTrace(System.err);
            System.exit(-1);
        }
        return null;
    }

    private static void findMethod(String name, List<FieldDescriptor> parameterTypes, String location, String caller,
            SyntaxNode node, QueryType queryType, Map<MethodInfo, DistanceTailPair> candidates, Set<ClassInfo> visitedInterfaces, boolean ignoreSynthetic) throws CompilerException, ByteCodeException, IOException, ResolveException {
        ClassFileRef locationRef = ClassFinder.instance().find(location);
        ClassFileRef callerRef = ClassFinder.instance().find(caller);
        String callerPackage = QIDTools.packageName(caller);
        checkRefs(locationRef, callerRef);
        ClassFile iter = locationRef.get();

        while (iter != null) {
            Collection<MethodInfo> methods = iter.getMethods(name);
            if (methods != null) {
                for (MethodInfo m : methods) {
                    if (!m.isStatic() && queryType == QueryType.STATIC_ONLY) {
                        //Not what we are looking for
                        continue;
                    } else if (m.isStatic() && (queryType == QueryType.NON_STATIC || queryType == QueryType.NON_VIRTUAL_NON_STATIC
                            || queryType == QueryType.NON_STATIC_EXACT)) {
                        continue;
                    } else if (m.isBridge() || (m.isSynthetic() && ignoreSynthetic)) {
                        //Ignore synthetic and bridge methods
                        continue;
                    } else if (m.isPrivate()) {
                        if (!m.getDefiningClass().getClassName().equals(caller)) {
                            //Ignore all private methods not owned by the caller
                            continue;
                        }
                    } else if (m.isProtected()) {
                        ClassFileRef owner = ClassFinder.instance().find(m.getDefiningClass().getClassName());
                        checkRefs(owner);
                        if (!DerivationTree.instance().isDerivated(callerRef, owner)) {
                            continue;
                        }
                    } else if (m.isPackagePrivate()) {
                        String packageName = QIDTools.packageName(m.getDefiningClass().getClassName());
                        if (!QIDTools.samePackage(callerPackage, packageName)) {
                            continue;
                        }
                    } else {
                        TeaJayAttribute tjattr = (TeaJayAttribute) m.getAttribute(TeaJayAttribute.NAME);
                        if (tjattr != null) {
                            if (tjattr.getAccessType() == TeaJayAttribute.AccessType.Hidden) {
                                continue;
                            }
                        }
                    }
                    if (queryType == QueryType.NON_STATIC_EXACT) {
                        if (m.getDescriptor().getParameters().equals(parameterTypes)) {
                            if (!candidates.containsKey(m)) {
                                candidates.put(m, new DistanceTailPair(0, 0));
                            }
                        }
                    } else {
                        Integer tail;
                        if ((tail = isCallable(m.getDescriptor().getParameters(), parameterTypes, m.isVarArgs())) != null) {
                            int distance = distance(m.getDescriptor().getParameters(), parameterTypes, m.isVarArgs());
                            if (!candidates.containsKey(m)) {
                                candidates.put(m, new DistanceTailPair(distance, tail));
                            }
                        }
                    }
                }
            }
            if (queryType != QueryType.STATIC_ONLY) {
                Collection<ClassInfo> interfaces = locationRef.get().getInterfaces();
                for (ClassInfo info : interfaces) {
                    if (!visitedInterfaces.contains(info)) {
                        visitedInterfaces.add(info);
                        findMethod(name, parameterTypes, info.getClassName(), caller, node, queryType, candidates, visitedInterfaces, ignoreSynthetic);
                    }
                }
            }
            if (queryType == QueryType.NON_VIRTUAL || queryType == QueryType.STATIC_ONLY
                    || queryType == QueryType.NON_VIRTUAL_NON_STATIC) {
                //No need to walk up the derivation tree for static methods and non virtual resolves
                iter = null;
            } else {
                ClassInfo superClass = iter.getSuper();
                if (superClass == null) {
                    iter = null;
                } else {
                    ClassFileRef ref = ClassFinder.instance().find(superClass.getClassName());
                    checkRefs(ref);
                    iter = ref.get();
                }
            }
        }
    }

    private static int distance(List<FieldDescriptor> fda, List<FieldDescriptor> fdb, boolean varArgs) throws CompilerException, ByteCodeException, ResolveException, IOException {
        if (fda.size() != fdb.size() && !varArgs) {
            throw new IllegalArgumentException("expecting lists of equal size for non vararg methods");
        } else if (fda.size() - 1 > fdb.size()) {
            throw new IllegalArgumentException("There was an error in isCallable()!");
        }
        Iterator<FieldDescriptor> ita = fda.iterator();
        Iterator<FieldDescriptor> itb = fdb.iterator();
        int dist = 0;
        while (ita.hasNext() && itb.hasNext()) {
            FieldDescriptor aNext = ita.next();
            if (!ita.hasNext() && varArgs) {
                if (aNext.isArray() && aNext.getDimension() == 1) {
                    FieldDescriptor leaf = aNext.getElementType();
                    while (itb.hasNext()) {
                        dist += distance(leaf, itb.next());
                    }
                } else {
                    dist += distance(aNext, itb.next());
                }
            } else {
                dist += distance(aNext, itb.next());
            }
        }

        return dist;
    }

    private static int distance(FieldDescriptor a, FieldDescriptor b) throws CompilerException, ByteCodeException, ResolveException, IOException {
        ClassFileRef refa = ClassFinder.instance().find(a.getClassInfo().getClassName());
        ClassFileRef refb = ClassFinder.instance().find(b.getClassInfo().getClassName());

        checkRefs(refa, refb);
        DerivationTree tree = DerivationTree.instance();
        if (refa.get().isInterface() && !refb.get().isInterface()) {
            return ifaceDistance(refb, refa, 1) + 1; //Punish interfaces
        } else if (refb.referencesNull()) {
            return 0;
        } else {
            int dist = tree.getDepth(refb) - tree.getDepth(refa);
            assert dist >= 0;
            return dist;
        }
    }

    private static int ifaceDistance(ClassFileRef type, ClassFileRef iface, int begin) throws CompilerException, ByteCodeException, ResolveException, IOException {
        assert type != null && iface != null && iface != type : "internal error";
        DerivationTree tree = DerivationTree.instance();
        int bestfit = Integer.MAX_VALUE;
        for (ClassInfo cancidate : type.get().getInterfaces()) {
            ClassFileRef cref = ClassFinder.instance().find(cancidate.getClassName());
            checkRefs(cref);
            if (tree.isAssignable(iface, cref)) {
                int tmp = tree.getDepth(cref) - tree.getDepth(iface) + begin;
                if (tmp < bestfit) {
                    bestfit = tmp;
                }
            }
        }
        ClassInfo superType = type.get().getSuper();
        if (superType != null && !superType.getClassName().equals("java/lang/Object")) {
            ClassFileRef superRef = ClassFinder.instance().find(superType.getClassName());
            checkRefs(superRef);
            if (tree.isAssignable(iface, superRef)) {
                int tmp = ifaceDistance(superRef, iface, begin + 1);
                if (tmp < bestfit) {
                    bestfit = tmp;
                }
            }
        }
        assert bestfit != Integer.MAX_VALUE && bestfit >= 0;
        return bestfit;
    }

    private static Integer isCallable(List<FieldDescriptor> fda, List<FieldDescriptor> fdb, boolean varArgs) throws CompilerException, ByteCodeException, IOException, ResolveException {
        int tail = 0;
        if (fda.size() == fdb.size()) {
            Iterator<FieldDescriptor> ita = fda.iterator();
            Iterator<FieldDescriptor> itb = fdb.iterator();
            while (ita.hasNext() && itb.hasNext()) {
                ClassFileRef refa = ClassFinder.instance().find(ita.next().getClassInfo().getClassName());
                ClassFileRef refb = ClassFinder.instance().find(itb.next().getClassInfo().getClassName());
                checkRefs(refa, refb);
                if (!DerivationTree.instance().isAssignable(refa, refb)) {
                    if (!itb.hasNext() && varArgs) {
                        if (refa.referencesArray()) {
                            FieldDescriptor descr = FieldDescriptor.fromString(refa.getQID());
                            if (descr.getDimension() == 1) {
                                ClassFileRef leaf = ClassFinder.instance().find(descr.getElementType().getClassInfo().getClassName());
                                checkRefs(leaf);
                                if (DerivationTree.instance().isAssignable(leaf, refb)) {
                                    tail = 1;
                                    continue;
                                }
                            }
                        }
                    }
                    return null;
                }
            }
            return tail;
        } else if (varArgs && (fda.size() < fdb.size() || fda.size() - 1 == fdb.size())) {
            Iterator<FieldDescriptor> ita = fda.iterator();
            Iterator<FieldDescriptor> itb = fdb.iterator();
            while (ita.hasNext() && itb.hasNext()) {
                ClassFileRef refa = ClassFinder.instance().find(ita.next().getClassInfo().getClassName());
                ClassFileRef refb = ClassFinder.instance().find(itb.next().getClassInfo().getClassName());
                checkRefs(refa, refb);
                if (!DerivationTree.instance().isAssignable(refa, refb)) {
                    if (ita.hasNext()) {
                        return null;
                    }
                    if (refa.referencesArray()) {
                        FieldDescriptor descr = FieldDescriptor.fromString(refa.getQID());
                        if (descr.getDimension() == 1) {
                            ClassFileRef leaf = ClassFinder.instance().find(descr.getElementType().getClassInfo().getClassName());
                            checkRefs(leaf);
                            do {
                                if (DerivationTree.instance().isAssignable(leaf, refb)) {
                                    tail++;
                                } else {
                                    return null;
                                }
                                if (itb.hasNext()) {
                                    refb = ClassFinder.instance().find(itb.next().getClassInfo().getClassName());
                                    checkRefs(refb);
                                } else {
                                    refb = null;
                                }
                            } while (refb != null);
                        } else {
                            return null;
                        }
                    } else {
                        return null;
                    }
                }
            }
            return tail;
        }
        return null;
    }

    private static void checkRefs(ClassFileRef... refs) throws Error {
        for (ClassFileRef ref : refs) {
            if (ref == null) {
                throw new Error("Encountered unknown type/class in stage 3.");
            }
        }
    }

    public static void main(String[] args) throws CompilerException {
        ClassFinder.instance().addPaths(FileIterator.BOOTPATH, FileIterator.CLASSPATH);
        Pair<MethodInfo, DistanceTailPair> print = MethodFinder.findMethod("print", "java/io/PrintStream", MethodFinder.class.getName().replace('.', '/'), null, QueryType.NON_STATIC_EXACT, FieldDescriptor.fromClass(int.class));
        System.err.println(print.value1);
    }
}
