/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.lang.java;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.util.Comparator;
import java.util.Iterator;
import teajay.lang.Closure;
import teajay.lang.Number;
import teajay.lang.TJComparable;
import teajay.lang.TJIterable;
import teajay.lang.TJIterator;
import teajay.util.TJSignature;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(accessType = TeaJayAttribute.AccessType.Final)
public final class JavaUtils {
    
    private JavaUtils() {
    }
    
    public static <T extends Comparable<? super T>> Number compare(T a, T b) {
        return Number.$valueOf(a.compareTo(b));
    }
    
    public static <T extends TJComparable<? super T>> Number compare(T a, T b) {
        return a.tjCompareTo(b);
    }
       
    @TJSignature("<T:Ljava/lang/Object;>(Ljava/util/Comparator<TT;>;)Lteajay/lang/Closure<Lteajay/lang/Number;TT;TT;>;")
    public static <T> Closure toTJComparator(final Comparator<T> c) {
        return new Closure() {

            @Override
            public Object execute(Object... params) {
                return Number.$valueOf(c.compare((T) params[0], (T) params[1]));
            }
        };
    }
    
    @TJSignature("<T:Ljava/lang/Object;>(Lteajay/lang/Closure<Lteajay/lang/Number;TT;TT;>;)Ljava/util/Comparator<TT;>;")
    public static <T> Comparator<T> toJavaComparator(final Closure c) {
        return new Comparator<T>() {

            @Override
            public int compare(T a, T b) {
                try {
                    return ((Number) c.execute(a, b)).$int_val();
                } catch (Throwable ex) {
                    throw new RuntimeException(ex);
                }
            }
        };
    }
    
    public static <T> Iterator<T> toJavaIterator(final TJIterator<T> c) {
        return new Iterator<T>() {

            @Override
            public boolean hasNext() {
                return c.hasNext();
            }

            @Override
            public T next() {
                return c.next();
            }

            @Override
            public void remove() {
                c.remove();
            }
        };
    }
    
    public static <T> TJIterator<T> toTJIterator(final Iterator<T> c) {
        return new TJIterator<T>() {

            @Override
            public Boolean hasNext() {
                return c.hasNext();
            }

            @Override
            public T next() {
                return c.next();
            }

            @Override
            public void remove() {
                c.remove();
            }
        };
    }
    
    public static <T> TJIterable<T> toTJIterable(final Iterator<T> c) {
        return new TJIterable<T>() {

            @Override
            public TJIterator<T> iterator() {
                return toTJIterator(c);
            }
        };
    }
    
    public static <T> Iterable<T> toJavaIterable(final TJIterator<T> c) {
        return new Iterable<T>() {

            @Override
            public Iterator<T> iterator() {
                return toJavaIterator(c);
            }
        };
    }
    
    @TJSignature("(Lteajay/lang/Closure<Ljava/lang/Void;>;Ljava/lang/String;)Ljava/lang/Thread;")
    public static Thread makeThread(final Closure c, String name) {
        return new Thread(new Runnable() {

            @Override
            public void run() {
                try {
                    c.execute();
                } catch (Throwable ex) {
                    Thread.currentThread().getUncaughtExceptionHandler().uncaughtException(Thread.currentThread(), ex);
                }
            }
            
        }, name);
    }
    
    @TJSignature("(Lteajay/lang/Closure<Ljava/lang/Void;>;)Ljava/lang/Thread;")
    public static Thread makeThread(final Closure c) {
        return new Thread(new Runnable() {

            @Override
            public void run() {
                try {
                    c.execute();
                } catch (Throwable ex) {
                    Thread.currentThread().getUncaughtExceptionHandler().uncaughtException(Thread.currentThread(), ex);
                }
            }
            
        });
    }
    
    public BigDecimal toBigDecimal(Number n, Number precision) {
        if (!precision.isInteger() || precision.isNegative()) {
            throw new IllegalArgumentException("precision must be a non-negative integer");
        }
        return n.$bigdec_val(new MathContext(precision.$int_val()));
    }
    
    public BigInteger toBigInteger(Number value) {
        return value.$bigint_val();
    }
}
