/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.lang.java;

import java.lang.reflect.Array;
import java.util.Arrays;
import teajay.lang.Number;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Interface zu java arrays. Type checking is deferred into runtime!
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(accessType = TeaJayAttribute.AccessType.Final)
public final class ReferenceArray<T> extends JavaArray<T> {

    private final Class<?> type;
    private final Object data;
    private final Number dimension;
    
    public ReferenceArray(Class<?> type, Number firstSize) {
        this(type, firstSize, (Number) null);
    }
    
    public ReferenceArray(Class<?> type, Number firstSize, Number... nextSizes) {
        if (nextSizes == null) {
            nextSizes = new Number[0];
        }
        Number[] size = new Number[nextSizes.length + 1];
        size[0] = firstSize;
        System.arraycopy(nextSizes, 0, size, 1, nextSizes.length);
        if (size.length <= 0 || size.length > 255) {
            throw new IllegalArgumentException("array dimension must be between 1 and 255 (inclusive)");
        }
        if (type.isPrimitive()) {
            throw new IllegalArgumentException("primitives are not supported by this array type");
        }
        if (type.isArray()) {
            throw new IllegalArgumentException("cannot have java arrays as component value");
        }
        this.type = type;
        this.dimension = Number.$valueOf(size.length);
        int[] sizes = new int[size.length];
        for (int i = 0; i < sizes.length; i++) {
            sizes[i] = size[i].$int_val();
        }
        this.data = Array.newInstance(type, sizes);
    }

    public ReferenceArray(Object obj) {
        Class<?> clazz = obj.getClass();
        if (!clazz.isArray()) {
            throw new IllegalArgumentException("expecting array argument");
        }
        this.data = obj;
        clazz = clazz.getComponentType();
        Class<?> last = null;
        int dim = 0;
        while (clazz != null) {
            dim++;
            last = clazz;
            clazz = clazz.getComponentType();
        }
        this.dimension = Number.$valueOf(dim);
        if (last != null && last.isPrimitive()) {
            throw new IllegalArgumentException("primitives are not supported by this array type");
        }
        this.type = last;
    }

    @Override
    public void set(int index, T val) {
        Object tmp = val;
        if (val instanceof ReferenceArray) {
            ReferenceArray referenceArray = (ReferenceArray) val;
            Class<?> componentType = referenceArray.getJavaArray().getClass();
            if (componentType != null && componentType.isArray()
                    && data.getClass().getComponentType().equals(componentType)) {
                tmp = referenceArray.getJavaArray();
            }
        }
        Array.set(data, index, tmp);
    }

    @Override
    public T get(int index) {
        Object obj = Array.get(data, index);
        if (obj == null) {
            return null;
        }
        Class<?> componentType = obj.getClass();
        if (componentType != null && componentType.isArray()
                && data.getClass().getComponentType().equals(componentType)) {
            obj = new ReferenceArray(obj);
        }
        return (T) obj;
    }
       
    @Override
    public Object getJavaArray() {
        return data;
    }

    @Override
    public Class<?> getArrayClass() {
        return data.getClass();
    }

    @Override
    public Number getDimension() {
        return dimension;
    }

    @Override
    public Class<?> getLeafElementType() {
        return type;
    }

    @Override
    public String toString() {
        return Arrays.deepToString((Object[]) data);
    }
    
    @Override
    public final Boolean operator_eq(Object obj) {
        return Arrays.deepEquals( (Object[]) getJavaArray(), (Object[]) obj);
    }
}
