/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.runtime;

import java.io.File;
import java.io.InputStream;
import java.io.PrintStream;
import java.io.StringReader;
import java.lang.invoke.MethodType;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import teajay.runtime.exceptions.TeaJayRuntimeError;
import static teajay.runtime.TypeManager.runtimePath;
import static teajay.runtime.TypeManager.mainConfig;
import teajay.util.ArrayUtils;
import teajay.util.TJType;
import teajay.util.io.ProgressToggler;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Looks up a given TeaJay type and calls the given Method by TeaJay's
 * invocation conventions.
 *
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(type = TeaJayAttribute.Type.Hidden)
public final class Run {

    private static final String version = "0.1";

    private enum ConstructorType {

        StringList,
        StringArray,
        Plain;
    }

    private Run() {
    }

    private static void t(PrintStream out) {
        out.print("\t");
    }

    private static void help(PrintStream out) {
        out.println("Usage:\ttj [options] <type name> <method name> [args]");
        out.println("or\ttj [options] -jar <jarfile> [args]");
        out.println("Options:");
        t(out);
        out.println("-cp <path>\tTeaJay's class path. This is where the runtime environment looks for types and implementations.");
        t(out);
        out.println("\tThe format is: <jar file or directory>{" + System.getProperty("path.separator", ":") + "<jar file or directory>}");
        t(out);
        out.println("-cfg <file>\t a config file that overrides implementations -> type mappings");
        t(out);
        out.println("-Ximpl <type>=<impl>\t set an implementation for <type>");
        t(out);
        out.println("-enable-rir\t enable runtime implementation replacement");
        t(out);
        out.println("-W \t\t enable warnings");
        t(out);
        out.println("-help\t\t show this help");
    }

    private static void version(PrintStream out) {
        out.println("TeaJay runtime environment version " + version);
        out.println("License: GPLv3 <http://www.gnu.org/licenses/gpl.html>");
        out.println("Creators of the TeaJay language: Michael Burzan and Michael Kußmann");
    }

    private static void init(Thread io) throws InterruptedException {
        io.setUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {
            @Override
            public void uncaughtException(Thread t, Throwable ex) {
                ex.printStackTrace(System.err);
                System.exit(-1);
            }
        });
        io.start();
        io.join(5 * 1000);
        ProgressToggler toggler = new ProgressToggler();
        if (io.isAlive()) {
            System.err.println("Try setting a less crowded classpath for faster startup");
        }
        while (io.isAlive()) {
            System.err.print("\rScanning for .tjt files " + toggler.toggle());
            io.join(1500);
        }
        if (toggler.hasToggled()) {
            System.err.println(" done.");
        }
    }

    public static void main(String[] args) throws Throwable {
        List<String> programArgs = new ArrayList<>();
        boolean encounteredProgramArgs = false;
        final Properties impls = new Properties();

        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-cp") && !encounteredProgramArgs) {
                if (args.length > i + 1) {
                    String path = System.getProperty(runtimePath);
                    if (path != null && !path.isEmpty()) {
                        path = path + System.getProperty("path.separator", ":");
                    } else {
                        path = "";
                    }
                    System.setProperty(runtimePath, path + args[i + 1]);
                } else {
                    throw new TeaJayRuntimeError("Expecting argument after -cp");
                }
                i++;
            } else if (args[i].equals("-cfg") && !encounteredProgramArgs) {
                if (args.length > i + 1) {
                    if (System.getProperty(mainConfig) != null) {
                        throw new TeaJayRuntimeError("may only specify one config file via -cfg");
                    }
                    System.setProperty(mainConfig, args[i + 1]);
                } else {
                    throw new TeaJayRuntimeError("Expecting argument after -cfg");
                }
                i++;
            } else if (args[i].equals("-jar") && !encounteredProgramArgs) {
                if (args.length > i + 1) {
                    File jarFile = new File(args[i + 1]);
                    if (!jarFile.canRead()) {
                        throw new TeaJayRuntimeError("cannot read jar file: " + jarFile);
                    }
                    if (!jarFile.getName().toLowerCase().endsWith(".jar")) {
                        throw new TeaJayRuntimeError("-jar expects jar file: " + jarFile);
                    }
                    System.setProperty("tj_jar", args[i + 1]);
                } else {
                    throw new TeaJayRuntimeError("Expecting argument after -jar");
                }
                encounteredProgramArgs = true; //no runtime arguments after -jar
                i++;
            } else if (args[i].equals("-enable-rir") && !encounteredProgramArgs) { //runtime implementation replacement
                Bootstraps.enableTypereset();
            } else if (args[i].equals("-Ximpl") && !encounteredProgramArgs) {
                if (args.length > i + 1) {
                    if (args[i + 1].contains("=")) {
                        try (StringReader reader = new StringReader(args[i + 1])) {
                            impls.load(reader);
                        }
                    } else {
                        throw new TeaJayRuntimeError("invalid usage of -Ximpl (" + args[i + 1] + ")");
                    }
                } else {
                    throw new TeaJayRuntimeError("expecting argument after -Ximpl");
                }
                i++;
            } else if ((args[i].equals("-help") || args[i].equals("--help")) && !encounteredProgramArgs) {
                help(System.out);
                System.exit(0);
            } else if ((args[i].equals("-version") || args[i].equals("--version")) && !encounteredProgramArgs) {
                version(System.out);
                System.exit(0);
            } else if (args[i].equals("-W") && !encounteredProgramArgs) {
                System.setProperty("tj_warn", "true");
            } else {
                encounteredProgramArgs = true;
                programArgs.add(args[i]);
            }
        }

        String jarFile = System.getProperty("tj_jar");
        if (jarFile == null) {
            //no jar file specified -> expecting type and method name
            if (programArgs.isEmpty()) {
                help(System.err);
                throw new TeaJayRuntimeError("must specify a type");
            }
            if (programArgs.size() == 1) {
                help(System.err);
                throw new TeaJayRuntimeError("must specify a method");
            }
        } else {
            //parse meta information on jar file
            try (ZipFile zipFile = new ZipFile(jarFile)) {
                ZipEntry entry = zipFile.getEntry("META-INF/TEAJAY.MF");
                if (entry == null) {
                    throw new TeaJayRuntimeError("jar file doesn't specify META-INF/TEAJAY.MF");
                }
                Properties p = new Properties();
                try (InputStream in = zipFile.getInputStream(entry);) {
                    p.load(in);
                }
                String mainClass = p.getProperty("Main-Class");
                String mainMethod = p.getProperty("Main-Method");
                if (mainClass == null) {
                    throw new TeaJayRuntimeError("jar file doesn't specify main type (Main-Class)");
                }
                if (mainMethod == null) {
                    throw new TeaJayRuntimeError("jar file doesn't specify main method (Main-Method)");
                }
                String jarClassPath = p.getProperty("Class-Path");
                if (jarClassPath != null) {
                    String[] cps = jarClassPath.split(" ");
                    for (String cp : cps) {
                        if (!cp.equals("")) {
                            String rp = System.getProperty(runtimePath, "");
                            System.setProperty(runtimePath, rp + (!rp.equals("") ? System.getProperty("path.separator") : "") + cp);
                        }
                    }
                }
                programArgs.add(0, mainClass);
                programArgs.add(1, mainMethod);
            }
        }

        args = programArgs.toArray(new String[programArgs.size()]);

        init(new Thread(new Runnable() {
            @Override
            public void run() {
                TypeManager.instance().putTypes(null, impls, "command line", true);
            }
        })); //Inititalize TypeManager in a separate thread

        TypeManager tm = TypeManager.instance();
        tm.setImplementationFinder(new PromptImplementationFinder());
        try {
            //Find a suitable construtor in the type
            Class<?> type = tm.getClassLoader().loadClass(args[0]);
            Class<?>[] params = new Class<?>[args.length - 2];
            Arrays.fill(params, String.class);
            ConstructorType ctype = ConstructorType.Plain;
            try {
                type.getConstructor(params);
            } catch (NoSuchMethodException ex) {
                try {
                    type.getConstructor(String[].class);
                    ctype = ConstructorType.StringArray;
                    params = new Class<?>[]{String[].class};
                } catch (NoSuchMethodException ex1) {
                    try {
                        Constructor<?> constr = type.getConstructor(teajay.lang.List.class);
                        checkConstructor(constr, type.getName() + " does not declare a compatible constructor", false);
                        ctype = ConstructorType.StringList;
                        params = new Class<?>[]{teajay.lang.List.class};
                    } catch (NoSuchMethodException ex2) {
                        throw new TeaJayRuntimeError("no such constructor " + MethodType.methodType(void.class, params));
                    }
                }
            }
            try {
                type.getMethod(args[1]);
            } catch (NoSuchMethodException ex) {
                throw new TeaJayRuntimeError("no such method " + args[1] + "()");
            }
            Class<?>[] newParams = new Class<?>[params.length + 1];
            newParams[0] = type;
            System.arraycopy(params, 0, newParams, 1, params.length);
            Class<?> impl = tm.getImpl(type);
            try {
                //Find a suitable construtor in the implementation
                Constructor<?> constr = impl.getConstructor(newParams);
                if (ctype == ConstructorType.StringList) {
                    checkConstructor(constr, impl.getName() + " does not conform with its type definition.", true);
                }
                Object[] typeArgs = new String[args.length - 2];
                System.arraycopy(args, 2, typeArgs, 0, args.length - 2);
                Object instance;
                switch (ctype) {
                    case StringList:
                        instance = constr.newInstance(null, ArrayUtils.toList(typeArgs));
                        break;
                    case Plain:
                        Object[] newTypeArgs = new Object[typeArgs.length + 1];
                        newTypeArgs[0] = null;
                        System.arraycopy(typeArgs, 0, newTypeArgs, 1, typeArgs.length);
                        instance = constr.newInstance(newTypeArgs);
                        break;
                    default: //StringArray
                        instance = constr.newInstance(null, (Object) typeArgs);
                }
                Method target = impl.getMethod(args[1]);
                final Object result = target.invoke(instance);
                if (!target.getReturnType().equals(void.class)) {
                    //print the result to console on exit
                    Runtime.getRuntime().addShutdownHook(new Thread(new Runnable() {
                        @Override
                        public void run() {
                            System.out.println(result);
                        }
                    }));
                }
            } catch (NoSuchMethodException ex) {
                throw new TeaJayRuntimeError("Type definition does not match implementation: " + impl.getName());
            } catch (InvocationTargetException ex) {
                //Hide the runtime environment from the stack-trace
                Throwable cause = ex.getCause();
                if (cause != null) {
                    StackTraceElement[] trace = cause.getStackTrace();
                    int last = -1;
                    for (int i = 0; i < trace.length; i++) {
                        if (trace[i].getClassName().equals(impl.getName()) && trace[i].getMethodName().equals(args[1])) {
                            last = i;
                        }
                    }
                    if (last == -1) {
                        throw cause;
                    } else {
                        cause.setStackTrace(Arrays.copyOfRange(trace, 0, last + 1));
                        throw cause;
                    }
                } else {
                    throw ex;
                }
            }
        } catch (ClassNotFoundException ex) {
            throw new TeaJayRuntimeError("No such type: " + args[0]);
        }
    }

    /**
     * Checks the generic signature of the constructor for beeing parameterized
     * with String.
     *
     * @param constr
     * @param msg
     * @param isImpl
     * @throws TeaJayRuntimeError if the cehck fails
     */
    private static void checkConstructor(Constructor<?> constr, String msg, boolean isImpl) throws TeaJayRuntimeError {
        Type[] types = constr.getGenericParameterTypes();
        boolean ok = false;
        int index = isImpl ? 1 : 0;
        if (types.length == index + 1 && types[index] instanceof ParameterizedType) {
            ParameterizedType ptype = (ParameterizedType) types[index];
            types = ptype.getActualTypeArguments();
            if (types.length == 1) {
                ok = types[0].equals(String.class);
            }
        }
        if (!ok) {
            throw new TeaJayRuntimeError(msg);
        }
    }
}
