/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.runtime;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Arrays;
import teajay.runtime.exceptions.TeaJayRuntimeError;
import teajay.util.TJType;
import teajay.util.tjattribute.TeaJayAttribute;

/**
 * Tools for working with TeaJay-Types in java.
 * This class allows to instantiate  TeaJay-Types and call static Methods on TeaJay-Types.
 * @author Michael Kußmann <michi@outx.de>
 */
@TJType(type = TeaJayAttribute.Type.Hidden)
public final class TJTypes {

    /**
     *
     * Instantiates a TeaJay type using TeaJay's runtime to find an
     * implementation for <tt> type </tt>.
     * This method uses the runtime types of the arguments to lookup a method.
     * @param <T>
     * @param type a TeaJay type
     * @param superCall
     * @param args
     * @return An instance of a type implementation of <tt>type</tt>.
     */
    public static <T> T instantiate(Class<T> type, boolean superCall, Object... args) {
        Class<?>[] paramTypes = new Class<?>[args.length];

        for (int i = 0; i < paramTypes.length; i++) {
            paramTypes[i] = args[i].getClass();
        }
        return instantiate(type, superCall, paramTypes, args);
    }
    
    public static <T> T instantiate(Class<T> type, boolean superCall, Class<?>[] paramTypes, Object... args) {
        if (paramTypes.length != args.length) {
            throw new IllegalArgumentException("The argument count does not match the count of supplied parameter types.");
        }
        TypeManager manager = TypeManager.instance();
        Class<? extends T> impl = manager.getImpl(type);
        if (!superCall) {
            Class<?>[] newParamTypes = new Class<?>[paramTypes.length + 1];
            newParamTypes[0] = type;
            System.arraycopy(paramTypes, 0, newParamTypes, 1, paramTypes.length);
            paramTypes = newParamTypes;
            
            Object[] tmp = new Object[args.length + 1];
            System.arraycopy(args, 0, tmp, 1, args.length);
            tmp[0] = null;
            args = tmp;
        }
        try {
            return impl.getConstructor(paramTypes).newInstance(args);
        } catch (NoSuchMethodException | InstantiationException | IllegalAccessException ex) {
            throw new TeaJayRuntimeError("No callable constructor " + type.getName() + ".<init> (" + Arrays.toString(paramTypes) + ") found.", ex);
        } catch (InvocationTargetException ex) {
            throw new TeaJayRuntimeError(ex.getCause());
        }
    }

    public static <T> T instantiate(Class<T> type, Object... args) {
        return instantiate(type, false, args);
    }
    
    public static <T> T instantiate(Class<T> type, Class<?>[] paramTypes, Object... args) {
        return instantiate(type, false, paramTypes, args);
    }

    public static Object invokeStatic(Class<?> type, String name, Object... args) {
        Class<?>[] paramTypes = new Class<?>[args.length];
        for (int i = 0; i < paramTypes.length; i++) {
            paramTypes[i] = args[i].getClass();
        }
        return invokeStatic(type, name, paramTypes, args);
    }

    public static Object invokeStatic(Class<?> type, String name, Class<?>[] paramTypes, Object... args) {
        if (paramTypes.length != args.length) {
            throw new IllegalArgumentException("The argument count does not match the count of supplied parameter types.");
        }
        try {
            type.getMethod(name, paramTypes);
            Class<?> impl = TypeManager.instance().getImpl(type);
            Method method = impl.getMethod(name, paramTypes);
            return method.invoke(null, args);
        } catch (NoSuchMethodException | SecurityException | IllegalAccessException | IllegalArgumentException ex) {
            throw new TeaJayRuntimeError("No callable method " + type.getName() + "." + name + "(" + Arrays.toString(paramTypes) + ") found.", ex);
        } catch (InvocationTargetException ex) {
            throw new TeaJayRuntimeError(ex.getCause());
        }
    }
}
