/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;

/**
 * This parser parses java number literals for use in TeaJay.
 * @author Michael Kußmann <michi@outx.de>
 */
public class JavaNumberParser {

    public static BigInteger parseBigInteger(String str, boolean sign) throws NumberFormatException {
        str = str.replace("_", "").toLowerCase();
        if (str.startsWith("0x")) {
            if (sign) {
                return new BigInteger(str.substring(2), 16);
            } else {
                return new BigInteger("-" + str.substring(2), 16);
            }
        } else if (str.startsWith("0b")) {
            if (sign) {
                return new BigInteger(str.substring(2), 2);
            } else {
                return new BigInteger("-" + str.substring(2), 2);
            }
        } else if (str.startsWith("0") && str.length() > 1) {
            if (sign) {
                return new BigInteger(str.substring(1), 8);
            } else {
                return new BigInteger("-" + str.substring(1), 8);
            }
        } else {
            if (sign) {
                return new BigInteger(str, 10);
            } else {
                return new BigInteger("-" + str, 10);
            }
        }
    }

    public static int parseIntegerLiteral(String str, boolean sign) throws NumberFormatException {
        str = str.replace("_", "").toLowerCase();

        if (str.startsWith("0x")) {
            if (sign) {
                return Integer.parseInt(str.substring(2), 16);
            } else {
                return Integer.parseInt("-" + str.substring(2), 16);
            }
        } else if (str.startsWith("0b")) {
            if (sign) {
                return Integer.parseInt(str.substring(2), 2);
            } else {
                return Integer.parseInt("-" + str.substring(2), 2);
            }
        } else if (str.startsWith("0") && str.length() > 1) {
            if (sign) {
                return Integer.parseInt(str.substring(1), 8);
            } else {
                return Integer.parseInt("-" + str.substring(1), 8);
            }
        } else {
            if (sign) {
                return Integer.parseInt(str, 10);
            } else {
                return Integer.parseInt("-" + str, 10);
            }
        }

    }

    public static long parseLongLiteral(String str, boolean sign) throws NumberFormatException {
        str = str.replace("_", "").toLowerCase();
        if (str.endsWith("l")) {
            str = str.substring(0, str.length() - 1);
        }

        if (str.startsWith("0x")) {
            if (sign) {
                return Long.parseLong(str.substring(2), 16);
            } else {
                return Long.parseLong("-" + str.substring(2), 16);
            }
        } else if (str.startsWith("0b")) {
            if (sign) {
                return Long.parseLong(str.substring(2), 2);
            } else {
                return Long.parseLong("-" + str.substring(2), 2);
            }
        } else if (str.startsWith("0") && str.length() > 1) {
            if (sign) {
                return Long.parseLong(str.substring(1), 8);
            } else {
                return Long.parseLong("-" + str.substring(1), 8);
            }
        } else {
            if (sign) {
                return Long.parseLong(str, 10);
            } else {
                return Long.parseLong("-" + str, 10);
            }
        }

    }

    public static float parseFloatLiteral(String str, boolean sign) throws NumberFormatException {
        str = str.replace("_", "").toLowerCase();
        if (str.endsWith("f")) {
            str = str.substring(0, str.length() - 1);
        }
        if (!sign) {
            str = "-" + str;
        }
        return Float.parseFloat(str);
    }

    public static double parseDoubleLiteral(String str, boolean sign) throws NumberFormatException {
        str = str.replace("_", "").toLowerCase();
        if (str.endsWith("d")) {
            str = str.substring(0, str.length() - 1);
        }
        if (!sign) {
            str = "-" + str;
        }
        return Double.parseDouble(str);
    }

    public static BigDecimal parseBigFloat(String str, boolean sign) {
        return parseBigFloat(str, sign, MathContext.UNLIMITED);
    }
    /**
     * Precondition: The lexer has recognized the string as beeing a floating
     * point literal.
     *
     * @param str
     * @param sign
     * @return
     */
    public static BigDecimal parseBigFloat(String str, boolean sign, MathContext context) {
        str = str.replace("_", "");
        char[] chars = str.toCharArray();
        int len;
        //Remove suffixes
        switch (chars[chars.length - 1]) {
            case 'f':
            case 'F':
            case 'd':
            case 'D':
                len = chars.length - 1;
                break;
            default:
                len = chars.length;
        }
        //check if the string starts with 0[xX]
        boolean hex = false;
        if (len > 2) {
            switch (chars[0]) {
                case '0':
                    switch (chars[1]) {
                        case 'x':
                        case 'X':
                            hex = true;
                    }
            }
        }
        if (hex) {
            int period = -1;
            int exp = -1;
            //find the starting indices of fractional and exponent part
            for (int i = 2; i < len; i++) {
                switch (chars[i]) {
                    case '.':
                        period = i;
                        continue;
                    case 'p':
                    case 'P':
                        exp = i;
                        break;
                    default:
                        continue;
                }
                break;
            }
            StringBuilder buf = new StringBuilder(len);
            if (period > -1) {
                buf.append(chars, 2, period - 2);
            }
            if (exp > -1) {
                if (period > -1) {
                    buf.append(chars, period + 1, exp - period - 1);
                } else {
                    buf.append(chars, 2, exp - 2);
                }
            } else {
                throw new NumberFormatException("Hexadecimal floating point literals need an exponent: " + str);
            }
            BigDecimal unscaled = new BigDecimal(new BigInteger(buf.toString(), 16), context);
            int exponent = Integer.parseInt(str.substring(exp + 1, len));
            //only singed 32 bit integer as exponent allowed
            BigDecimal p = new BigDecimal(BigInteger.ONE.shiftLeft(Math.abs(exponent)), context);
            if (exponent < 0) {
                p = BigDecimal.ONE.divide(p, context);
            }
            int hexScale;
            if (period > -1) {
                hexScale = exp - period - 1;
            } else {
                hexScale = 0;
            }
            BigDecimal scale = new BigDecimal(BigInteger.ONE.shiftLeft(hexScale * 4), context); //16^hexScale
            //scale the unscaled value
            BigDecimal scaled = unscaled.multiply(BigDecimal.ONE.divide(scale), context).multiply(p, context); //unscaled * 16^-hexScale * 2^exponent
            if (!sign) {
                scaled = scaled.negate(context);
            }
            return scaled;
        } else {
            if (!sign) {
                str = "-" + str;
                return new BigDecimal(str.toCharArray(), 0, len + 1, context);
            } else {
                return new BigDecimal(chars, 0, len, context);
            }
        }
    }

    public static void main(String[] args) {
        String str = "0x2c3.afb3p2";
        System.err.println(parseDoubleLiteral(str, true));
        System.err.println(parseBigFloat(str, true, MathContext.UNLIMITED).doubleValue());
    }
}
