/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.io;

import teajay.runtime.exceptions.DuplicateQualifiedIdentifierException;
import java.io.Closeable;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * TODO: Lazy resolving of ressources: Only search for qids on demand,
 * that will reduce startup time in crowded working directories.
 * @author Michael Kußmann <michi@outx.de>
 */
abstract class AbstractResourceFinder<T> {

    protected static class QID {

        public final String javaPackage;
        public final String simpleName;

        protected QID(String javaPackage, String simpleName) {
            this.javaPackage = javaPackage;
            this.simpleName = simpleName;
        }

        @Override
        public String toString() {
            return javaPackage + "/" + simpleName;
        }
    }
    private final Collection<Closeable> iterators = Collections.synchronizedCollection(new LinkedList<Closeable>());
    private final Map<String, Map<String, T>> fileCache = new HashMap<>(1024);
    
    protected AbstractResourceFinder() {
    }
    
    /**
     * Adds new paths to the finder.
     *
     * @param paths an array containing path strings in the format PATH:PATH:...
     */
    protected synchronized void addPaths(FilesVisitor visitor, boolean scanJar, String... paths) throws DuplicateQualifiedIdentifierException {
        FileIterator it = FileIterator.fromPathStrings(paths);
        iterators.add(it);
        it.scan(visitor, scanJar);
    }

    protected static QID split(String qID) {
        int index = qID.lastIndexOf("/");
        String javaPackage;
        String simpleName;
        if (index < 0) {
            javaPackage = null;
            simpleName = qID;
        } else {
            javaPackage = qID.substring(0, index);
            simpleName = qID.substring(index + 1);
        }
        return new QID(javaPackage, simpleName);
    }

    protected void putRef(String javaPackage, String simpleName, T ref, boolean force) throws DuplicateQualifiedIdentifierException {
        Map<String, T> map = fileCache.get(javaPackage);
        if (map == null) {
            map = Collections.synchronizedMap(new HashMap<String, T>());
            fileCache.put(javaPackage, map);
        }
        if (!force && map.containsKey(simpleName)) {
            throw new DuplicateQualifiedIdentifierException(javaPackage, simpleName);
        } else {
            map.put(simpleName, ref);
        }
    }

    public synchronized T find(String qID) {
        QID q = split(qID);
        return find(q.javaPackage, q.simpleName);
    }
    
    public synchronized T find(String javaPackage, String simpleName) {
        Map<String, T> map = fileCache.get(javaPackage);
        return map != null ? map.get(simpleName) : null;
    }

    public synchronized Map<String, T> findAll(String javaPackage) {
        Map<String, T> map = fileCache.get(javaPackage);
        if (map != null) {
            return Collections.unmodifiableMap(map);
        } else {
            return null;
        }
    }

    public int countEntries() {
        int count = 0;
        for (Map<?, ?> m : fileCache.values()) {
            count += m.size();
        }
        return count;
    }
    /**
     * 
     * @throws IOException 
     */
    protected void destroy() throws IOException {
        for (Closeable c : iterators) {
            try {
                c.close();
            } catch (IOException ex) {
            }
        }
    }
}
