/*
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
    
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package teajay.util.io;

import teajay.runtime.exceptions.ResolveException;
import java.io.DataInputStream;
import java.io.IOException;
import java.lang.ref.SoftReference;
import teajay.bytecode.ClassFile;
import static teajay.bytecode.ClassFile.ReadMode.*;
import teajay.bytecode.exceptions.ByteCodeException;

/**
 *
 * @author Michael Kußmann <michi@outx.de>
 */
public abstract class ClassFileRef {
    public abstract ClassFile get() throws ByteCodeException, IOException, ResolveException;
    public boolean referencesPrimitive() {
        return false;
    }
    
    public boolean referencesArray() {
        return false;
    }
    
    public boolean referencesNull() {
        return false;
    }
    
    public abstract String getQID();
}

abstract class LazyClassFileRef extends ClassFileRef {

    private final Input input;
    private final String qID;

    protected LazyClassFileRef(Input input, String qID) {
        if (input == null || qID == null) {
            throw new NullPointerException(input == null ? "input" : "qID");
        }
        this.input = input;
        this.qID = qID;
    }
    
    protected ClassFile read() throws IOException, ByteCodeException, ResolveException {
        try (DataInputStream in = new DataInputStream(input.openStream())) {
            ClassFile file = ClassFile.read(in, ForgetConstantPool, All);
            if (!qID.equals(file.getThis().getClassName())) {
                throw new ResolveException("class's package name does not match its directory structure: dir=" + qID + ", package=" + file.getThis().getClassName());
            }
            return file;
        }
    }
    
    @Override
    public String toString() {
        return input.toString();
    }
    
    @Override
    public String getQID() {
        return qID;
    }
}
class SoftClassFileRef extends LazyClassFileRef {
    private SoftReference<ClassFile> classRef;
    

    SoftClassFileRef(Input input, String qID) {
        super(input, qID);
    }

    @Override
    public synchronized ClassFile get() throws IOException, ByteCodeException, ResolveException {
        if (classRef == null) {
            ClassFile cl = read();
            this.classRef = new SoftReference<>(cl);
            return cl;
        }

        ClassFile cl = classRef.get();
        if (cl == null) {
            cl = read();
            this.classRef = new SoftReference<>(cl);
        }
        return cl;
    }
}

class StrongClassFileRef extends LazyClassFileRef {

    private ClassFile classFile;

    StrongClassFileRef(Input input, String qID) {
        super(input, qID);
    }

    @Override
    public synchronized ClassFile get() throws ByteCodeException, IOException, ResolveException {
        if (classFile == null) {
            this.classFile = read();
        }
        return classFile;
    }
}


class ArrayDummyClassFileRef extends ClassFileRef {
    private final String qID;
    private final ClassFile dummy;
    
    ArrayDummyClassFileRef(String qID, ClassFile dummy) {
        this.qID = qID;
        this.dummy = dummy;
    }

    @Override
    public ClassFile get() throws ByteCodeException, IOException, ResolveException {
        return dummy;
    }

    @Override
    public String getQID() {
        return qID;
    }
    
    @Override
    public boolean referencesArray() {
        return true;
    }
}

class DirectClassFileRef extends ClassFileRef {
    private final ClassFile classFile;
    private final boolean isPrimitive;
    private final boolean isNull;
    
    DirectClassFileRef(ClassFile classFile) {
        if (classFile == null) {
            throw new NullPointerException("classFile");
        }
        this.classFile = classFile;
        this.isPrimitive = classFile.getThis().isPrimitive();
        this.isNull = classFile.getThis().getClassName().equals("null");
    }
    
    @Override
    public ClassFile get() {
        return classFile;
    }
    
    @Override
    public boolean referencesPrimitive() {
        return isPrimitive;
    }
    
    @Override
    public boolean referencesNull() {
        return isNull;
    }
    
    @Override
    public String toString() {
        return "Direct proxy for " + classFile.getThis().getClassName();
    }
    
    @Override
    public String getQID() {
        return classFile.getThis().getClassName();
    }
    
}